import sys
from collections import ChainMap
import threading
import numpy as np
import lmfit
from collections import deque

import matplotlib

matplotlib.use("Qt5Agg")
import matplotlib.pyplot as plt
from bluesky.callbacks.mpl_plotting import LivePlot, LiveFitPlot
from bluesky.callbacks import LiveFit
from bluesky.callbacks.core import get_obj_fields
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT

from PySide6.QtWidgets import (
    QWidget,
    QGridLayout,
    QApplication,
    QPushButton,
    QTableWidgetItem,
    QColorDialog,
    QComboBox,
    QLabel,
    QLineEdit,
)
from PySide6.QtCore import Signal, QObject, Qt, QCoreApplication
from PySide6.QtGui import QIcon

from nomad_camels.gui.plot_options import Ui_Plot_Options
from nomad_camels.utility.fit_variable_renaming import replace_name
from nomad_camels.bluesky_handling.evaluation_helper import Evaluator
from ophyd import SignalRO, Device, Component, BlueskyInterface, Kind
from bluesky import plan_stubs as bps

from nomad_camels.utility.plot_placement import place_widget

from importlib import resources
from nomad_camels import graphics
from nomad_camels.ui_widgets.warn_popup import WarnPopup

stdCols = plt.rcParams["axes.prop_cycle"].by_key()["color"]

dark_mode = False


def activate_dark_mode():
    """Changes the plot-style to dark-mode."""
    global dark_mode
    dark_mode = True
    plt.style.use("dark_background")


class MPLwidget(FigureCanvasQTAgg):
    """Custom QT widget for displaying matplotlib plots.

    This class inherits from matplotlib's FigureCanvasQTAgg to create a custom
    QT widget for displaying matplotlib plots. In the init method, a new figure
    and axes are created using matplotlib.pyplot.subplots(). The grid is then
    displayed on the axes.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self):
        fig, ax = plt.subplots()
        self.axes = ax
        self.axes.grid()
        super().__init__(fig)


class PlotWidget(QWidget):
    """Class for creating a plot widget.

    Parameters
    ----------
    x_name : str
        The name of the x-axis variable
    y_names : Union[str, Tuple[str]]
        The name(s) of the y-axis variable(s)
    legend_keys : List[str], optional
        The keys for the legend, by default None
    xlim : Tuple[float, float], optional
        passed to Axes.set_xlim
    ylim : Tuple[float, float], optional
        passed to Axes.set_ylim
    epoch : {'run', 'unix'}, optional
        If 'run' t=0 is the time recorded in the RunStart document. If 'unix',
        t=0 is 1 Jan 1970 ("the UNIX epoch"). Default is 'run'.
    parent : QWidget, optional
        The parent widget, by default None
    namespace : Optional[Mapping[str, Any]]
        The namespace to use for the `Evaluator`, by default None
    ylabel : str, optional
        The y-axis label, by default ''
    xlabel : str, optional
        The x-axis label, by default ''
    title : str, optional
        The title of the plot, by default ''
    stream_name : str, optional
        The name of the stream to be used for the plot. Default is 'primary'
    fits : List[Dict[str, Union[str, bool, List[str], Tuple[float, float], Dict[str, Union[str, float]]]]], optional
        The fits for the plot, by default None
    do_plot : bool, optional
        Whether to show the plot, by default True
    **kwargs : Any, optional
        Additional keyword arguments to pass to `MultiLivePlot`

    Returns
    -------

    Attributes
    ----------
    ax : Axes
        The matplotlib axes of the plot
    x_name : str
        The name of the x-axis variable
    y_names : List[str]
        The name(s) of the y-axis variable(s)
    stream_name : str
        The name of the stream
    fits : List[Dict[str, Union[str, bool, List[str], Tuple[float, float], Dict[str, Union[str, float]]]]]
        The fits for the plot as they come from the fit/plot definer.
    liveFits : List[LiveFit_Eva]
        The live fit objects for the plot, handled by the liveFitPlots
    liveFitPlots : List[Fit_Plot_No_Init_Guess]
        The live fit plots for the plot, used to display the fits
    livePlot : MultiLivePlot
        The live plot, using the canvas etc.
    toolbar : NavigationToolbar2QT
        The toolbar for the plot
    pushButton_show_options : QPushButton
        The push button to show the plot options
    pushButton_autoscale : QPushButton
        The push button to autoscale the plot
    plot_options : Plot_Options
        The options widget for the plot
    options_open : bool
        Whether the options are currently open
    """

    closing = Signal()

    def __init__(
        self,
        x_name,
        y_names,
        *,
        legend_keys=None,
        xlim=None,
        ylim=None,
        epoch="run",
        parent=None,
        namespace=None,
        ylabel="",
        xlabel="",
        title="",
        stream_name="primary",
        fits=None,
        do_plot=True,
        multi_stream=False,
        y_axes=None,
        logX=False,
        logY=False,
        logY2=False,
        maxlen=np.inf,
        **kwargs,
    ):
        super().__init__(parent=parent)
        canvas = MPLwidget()
        if isinstance(y_names, str):
            y_names = [y_names]
        self.ax = canvas.axes
        self.x_name = x_name
        self.y_names = y_names
        self.stream_name = stream_name
        self.fits = fits or []
        self.liveFits = []
        self.liveFitPlots = []
        eva = Evaluator(namespace=namespace)
        self.ax2 = self.ax.twinx() if y_axes and 2 in y_axes.values() else None
        for fit in self.fits:
            if fit["use_custom_func"]:
                model = lmfit.models.ExpressionModel(fit["custom_func"])
                label = "custom"
            else:
                model = lmfit.models.lmfit_models[fit["predef_func"]]()
                label = fit["predef_func"]
            if fit["guess_params"]:
                init_guess = None
            else:
                init_guess = {}
                for i, param in enumerate(fit["initial_params"]["name"]):
                    init_guess[param] = fit["initial_params"]["initial value"][i]
            upper = {}
            lower = {}
            for i, param in enumerate(fit["initial_params"]["name"]):
                upper[param] = fit["initial_params"]["upper bound"][i] or np.inf
                lower[param] = fit["initial_params"]["lower bound"][i] or -np.inf
            params = model.make_params()
            if init_guess:
                for i, param in enumerate(fit["initial_params"]["name"]):
                    params[param].set(
                        init_guess[param], min=lower[param], max=upper[param]
                    )
            else:
                for i, param in enumerate(fit["initial_params"]["name"]):
                    params[param].set(min=lower[param], max=upper[param])
            name = f'{label}_{fit["y"]}_v_{fit["x"]}'
            name = replace_name(name)
            add_data = fit["additional_data"] or {}
            livefit = LiveFit_Eva(
                model,
                fit["y"],
                {"x": fit["x"]},
                eva,
                init_guess,
                name=name,
                additional_data=add_data,
                params=params,
                stream_name=stream_name,
            )
            self.liveFits.append(livefit)
            if y_axes and y_axes[fit["y"]] == 2:
                ax = self.ax2
                ax_is2 = True
            else:
                ax = self.ax
                ax_is2 = False
            if fit["y"] in y_names:
                col = stdCols[y_names.index(fit["y"])]
            else:
                col = "black"
            self.liveFitPlots.append(
                Fit_Plot_No_Init_Guess(
                    livefit,
                    ax=ax,
                    legend_keys=[name],
                    ax_is2=ax_is2,
                    color=col,
                    display_values=fit["display_values"],
                )
            )
        self.livePlot = MultiLivePlot(
            y_names,
            x_name,
            legend_keys=legend_keys,
            xlim=xlim,
            ylim=ylim,
            epoch=epoch,
            ax=canvas.axes,
            evaluator=eva,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
            stream_name=stream_name,
            do_plot=do_plot,
            fitPlots=self.liveFitPlots,
            y_axes=y_axes,
            multi_stream=multi_stream,
            ax2=self.ax2,
            **kwargs,
        )
        self.livePlot.new_data.connect(self.show)
        self.toolbar = NavigationToolbar2QT(canvas, self)

        self.pushButton_show_options = QPushButton("Show Options")
        self.pushButton_show_options.clicked.connect(self.show_options)
        self.pushButton_autoscale = QPushButton("Autoscale")
        self.pushButton_autoscale.clicked.connect(self.autoscale)
        self.pushButton_clear = QPushButton("Clear Plot")
        self.pushButton_clear.clicked.connect(self.clear_plot)
        self.plot_options = Plot_Options(self, self.ax, self.livePlot, self.ax2)
        label_n_data = QLabel("# data points:")
        self.lineEdit_n_data = QLineEdit(str(maxlen))
        self.lineEdit_n_data.returnPressed.connect(self.change_maxlen)

        layout = QGridLayout()
        layout.addWidget(canvas, 0, 1, 1, 6)
        layout.addWidget(self.toolbar, 1, 4)
        layout.addWidget(self.pushButton_show_options, 1, 1)
        layout.addWidget(self.pushButton_autoscale, 1, 2)
        layout.addWidget(self.pushButton_clear, 1, 3)
        layout.addWidget(label_n_data, 1, 5)
        layout.addWidget(self.lineEdit_n_data, 1, 6)
        layout.addWidget(self.plot_options, 0, 0, 2, 1)
        self.setLayout(layout)

        self.setWindowTitle(title or f"{x_name} vs. {y_names[0]}")
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))

        self.plot_options.hide()
        self.options_open = False
        if do_plot:
            self.show()
            self.plot_options.checkBox_log_x.setChecked(logX)
            self.plot_options.checkBox_log_y.setChecked(logY)
            self.plot_options.checkBox_log_y2.setChecked(logY2)
            self.plot_options.set_log()
        place_widget(self)
        self.change_maxlen()

    def change_maxlen(self):
        """ """
        text = self.lineEdit_n_data.text()
        if (
            not text
            or text == "None"
            or text == "none"
            or text == "inf"
            or text == "np.inf"
        ):
            maxlen = np.inf
        else:
            try:
                maxlen = int(text)
            except:
                return
        self.livePlot.change_maxlen(maxlen)

    def clear_plot(self):
        """Clear the plot by removing the data from the plot and clearing all
        fit plots.

        Parameters
        ----------

        Returns
        -------

        """
        self.livePlot.clear_plot()
        for fit_plot in self.liveFitPlots:
            fit_plot.clear_plot()

    def autoscale(self):
        """Autoscale the plot's x and y axis."""
        self.ax.autoscale()
        self.ax.figure.canvas.draw_idle()

    def show_options(self):
        """Show or hide the options for the plot.
        Toggles between 'Show Options' and 'Hide Options' on the button press.

        Parameters
        ----------

        Returns
        -------

        """
        if self.options_open:
            self.pushButton_show_options.setText("Show Options")
            self.options_open = False
            self.plot_options.hide()
        else:
            self.pushButton_show_options.setText("Hide Options")
            self.options_open = True
            self.plot_options.show()
            self.plot_options.setup_table()
        self.adjustSize()

    def closeEvent(self, a0):
        """
        Overwrite the closeEvent to emit the closing signal before closing the
        window.

        Parameters
        ----------
        a0 :


        Returns
        -------

        """
        self.closing.emit()
        super().closeEvent(a0)


class LiveFit_Eva(LiveFit):
    """LiveFit_Eva is a subclass of LiveFit that adds the ability to evaluate the
    independent variables before fitting. It uses the given evaluator for that.

    Parameters
    ----------

    Returns
    -------


    """

    def __init__(
        self,
        model,
        y,
        independent_vars,
        evaluator,
        init_guess=None,
        additional_data=None,
        *,
        name="",
        params=None,
        stream_name="primary",
    ):
        super().__init__(
            model=model, y=y, independent_vars=independent_vars, init_guess=init_guess
        )
        self.eva = evaluator
        self.name = f"{name}_{stream_name}"
        self.params = params
        name = replace_name(name)
        # self.ophyd_fit = Fit_Ophyd(name, name=name, params=params,
        #                            parent_fit=self)
        self.stream_name = stream_name
        self.timestamp = None
        self.parent_plot = None
        self.__stale = True
        self.ready_to_read = False
        self.results = {}
        if isinstance(additional_data, list):
            self.additional_data = {}
            for d in additional_data:
                self.additional_data[d] = []
        else:
            self.additional_data = additional_data or {}
        self.read_ready = Fit_Signal(f"{self.name}_read_ready")

    def _reset(self):
        """Resets the fit, also sets the `ready_to_read` to False."""
        super()._reset()
        self.ready_to_read = False

    def event(self, doc):
        """Handles new events received by the fit. Evaluates the independent variables using the evaluator
        and updates the fit with the new data.

        Parameters
        ----------
        doc :


        Returns
        -------


        """
        idv = {}
        for k, v in self.independent_vars.items():
            try:
                new_x = doc["data"][v]
            except KeyError:
                if v in ("time", "seq_num"):
                    new_x = doc[v]
                else:
                    if not self.eva.is_to_date(doc["time"]):
                        self.eva.event(doc)
                    new_x = self.eva.eval(v)
            idv[k] = new_x

        try:
            y = doc["data"][self.y]
        except KeyError:
            if not self.eva.is_to_date(doc["time"]):
                self.eva.event(doc)
            y = self.eva.eval(self.y)
        # Always stash the data for the next time the fit is updated.
        self.update_caches(y, idv)
        self.timestamp = doc["time"]
        self.__stale = True

    def get_ready(self):
        """This function is used to set the ready_to_read attribute to True."""
        self.ready_to_read = True

    def update_fit(self):
        """Update the fit by evaluating the model with the current data.

        This method updates the fit by evaluating the current data using the
        model defined by the user. The method uses the `params` attribute, if
        available, to pass on the parameters to the fitting method. If the fit
        is not stale (i.e. no new data has arrived since the last update) the
        method will return without doing anything. Otherwise, the method updates
        the `result` attribute with the fit result and the `ophyd_fit` with the
        updated data. It also calls the `fit_has_result` method on the parent
        plot.
        Before doing anything, it will wait for `ready_to_read` to be True.

        Parameters
        ----------

        Returns
        -------

        """
        while not self.ready_to_read:
            yield from bps.sleep(0.1)
        if not self.__stale:
            return None
        kwargs = {}
        kwargs.update(self.independent_vars_data)
        kwargs.update(self.init_guess)
        try:
            if self.params:
                self.result = self.model.fit(self.ydata, params=self.params, **kwargs)
            else:
                self.result = self.model.fit(self.ydata, **kwargs)
        except Exception as e:
            print(f"Error in fit {self.name}: {e}")
            return None
        self.results[f"{self.timestamp}"] = self.result
        for d in self.additional_data:
            self.additional_data[d].append(self.eva.eval(d))
        self.__stale = False
        # self.ophyd_fit.update_data(self.result, self.timestamp)
        self.parent_plot.fit_has_result()


class Fit_Signal(SignalRO):
    """A subclass of ophyd.SignalRO for storing fit results

    This class is a subclass of ophyd.SignalRO, which is used to store the
    results of a fit.
    It has an additional method `update_data` that updates the readback value
    and timestamp metadata of the signal.

    Parameters
    ----------

    Returns
    -------


    """

    def __init__(
        self,
        name,
        value=0.0,
        timestamp=None,
        parent=None,
        labels=None,
        kind="hinted",
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
    ):
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )

    def update_data(self, result, timestamp):
        """Updates the readback value and timestamp metadata of the signal.
        Called by the parent's `update_data` function.

        Parameters
        ----------
        result :

        timestamp :


        Returns
        -------

        """
        self._readback = result
        self._metadata["timestamp"] = timestamp


class Fit_Ophyd(Device):
    """A device that extends the functionality of the Device class from Ophyd.
    It is included in the LiveFit_Eva class.

    Parameters
    ----------

    Returns
    -------


    """

    a = Component(Fit_Signal, name="a")
    b = Component(Fit_Signal, name="b")
    c = Component(Fit_Signal, name="c")
    d = Component(Fit_Signal, name="d")
    e = Component(Fit_Signal, name="e")
    f = Component(Fit_Signal, name="f")
    g = Component(Fit_Signal, name="g")
    h = Component(Fit_Signal, name="h")
    i = Component(Fit_Signal, name="i")
    j = Component(Fit_Signal, name="j")
    k = Component(Fit_Signal, name="k")
    l = Component(Fit_Signal, name="l")
    m = Component(Fit_Signal, name="m")
    n = Component(Fit_Signal, name="n")
    o = Component(Fit_Signal, name="o")
    p = Component(Fit_Signal, name="p")
    q = Component(Fit_Signal, name="q")
    r = Component(Fit_Signal, name="r")
    covar = Component(Fit_Signal, name="covar")
    read_ready = Component(Fit_Signal, name="read_ready")

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        params=None,
        parent_fit=None,
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.params = params.keys()
        self.order = [
            self.a,
            self.b,
            self.c,
            self.d,
            self.e,
            self.f,
            self.g,
            self.h,
            self.i,
            self.j,
            self.k,
            self.l,
            self.m,
            self.n,
            self.o,
            self.p,
            self.q,
            self.r,
        ]
        self.used_comps = [self.covar]
        for i, p in enumerate(self.params):
            self.used_comps.append(self.order[i])
        self.parent_fit = parent_fit

    def update_data(self, result, timestamp):
        """Update the data of all the components.
        For each fit parameter, the corresponding component is updated.

        Parameters
        ----------
        result : lmfit.model.ModelResult
            The result of the fit.

        timestamp : float
            The timestamp of the fit. Used to update the timestamp metadata of
            the components.

        Returns
        -------


        """
        self.used_comps = [self.covar]
        for i, comp in enumerate(self.params):
            if comp in result.best_values:
                self.order[i].update_data(result.best_values[comp], timestamp)
                self.used_comps.append(self.order[i])
                self.order[i].name = f"{self.name}_{comp}"
        if result is not None and result.covar is not None:
            self.covar.update_data(result.covar, timestamp)
        else:
            self.covar.update_data(None, timestamp)

    def read(self):
        """Overwrites the `read` method from `Device`.
        Stops reading, as soon as the number of parameters is reached.
        """
        res = BlueskyInterface.read(self)
        i = 1
        for _, component in self._get_components_of_kind(Kind.normal):
            res.update(component.read())
            i += 1
            if i > len(self.params):
                break
        return res


class Fit_Plot_No_Init_Guess(LiveFitPlot):
    """A subclass of LiveFitPlot that doesn't plot the initial guess for the fit."""

    def __init__(
        self,
        livefit,
        *,
        num_points=100,
        legend_keys=None,
        xlim=None,
        ylim=None,
        ax=None,
        ax_is2=False,
        display_values=False,
        **kwargs,
    ):
        # super().__init__(livefit, num_points=num_points, legend_keys=legend_keys,
        #                  xlim=xlim, ylim=ylim, ax=ax, **kwargs)
        self.__setup_lock = threading.Lock()
        self.__setup_event = threading.Event()
        self.display_values = display_values
        if len(livefit.independent_vars) != 1:
            raise NotImplementedError(
                "LiveFitPlot supports models with one " "independent variable only."
            )

        def setup():
            # Run this code in start() so that it runs on the correct thread.
            nonlocal legend_keys, xlim, ylim, ax, kwargs, livefit
            with self.__setup_lock:
                if self.__setup_event.is_set():
                    return
                self.__setup_event.set()
            self.ax = ax
            # if legend_keys is None:
            #     legend_keys = []
            # self.legend_keys = ['scan_id'] + legend_keys
            (x,) = livefit.independent_vars.values()
            if x is not None:
                self.x, *others = get_obj_fields([x])
            else:
                self.x = "seq_num"
            y = livefit.y
            self.y, *others = get_obj_fields([y])
            if xlim is not None:
                self.ax.set_xlim(*xlim)
            if ylim is not None:
                self.ax.set_ylim(*ylim)
            # self.ax.margins(.1)
            self.kwargs = kwargs
            self.lines = []
            # self.legend = None
            # self.legend_title = " :: ".join([name for name in self.legend_keys])
            # self._epoch_offset = None  # used if x == 'time'
            # self._epoch = epoch

        (self.__x_key,) = livefit.independent_vars.keys()  # this never changes
        # x, = livefit.independent_vars.values()  # this may change
        self.num_points = num_points
        self._livefit = livefit
        self._xlim = xlim
        self._has_been_run = False
        livefit.parent_plot = self
        self.x_data = []
        self.y_data = []
        self.current_line = None
        self.x = None
        self.ax_is2 = ax_is2
        self.__setup = setup
        self.line_position = None
        self.text_objects = []

    def start(self, doc):
        """Overwrites the `start` method of LiveFitPlot to not display the
        init_guess_line.

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        # LivePlot.start(self, doc)
        self.__setup()
        # The doc is not used; we just use the signal that a new run began.
        # self._epoch_offset = doc['time']  # used if self.x == 'time'
        self.x_data, self.y_data = [], []
        (self.current_line,) = self.ax.plot([], [], **self.kwargs)
        self.lines.append(self.current_line)
        # legend = self.ax.legend(loc=0, title=self.legend_title)
        # try:
        #     # matplotlib v3.x
        #     self.legend = legend.set_draggable(True)
        # except AttributeError:
        #     # matplotlib v2.x (warns in 3.x)
        #     self.legend = legend.draggable(True)
        self.livefit.start(doc)
        (self.x,) = self.livefit.independent_vars.keys()  # in case it changed
        # Put fit above other lines (default 2) but below text (default 3).
        [line.set_zorder(2.5) for line in self.lines]
        # if self.legend_keys:
        #     self.current_line.set_label(self.legend_keys[-1])
        # legend = self.ax.legend(loc=0, title='')
        # if not self.ax_is2:
        #     try:
        #         legend.set_draggable(True)
        #     except AttributeError:
        #         legend.draggable(True)
        # else:
        #     self.ax.get_legend().remove()

    def get_ready(self):
        """Passes the command to the `_livefit`"""
        self._livefit.get_ready()

    def event(self, doc):
        """Passes the event to the `livefit`

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        self.livefit.event(doc)

    def fit_has_result(self):
        """If the `livefit` has a result, a resulting line is calculated, and the
        plot is updated.
        Called by LiveFit_Eva.update_fit().

        Parameters
        ----------

        Returns
        -------

        """
        if self.livefit.result is not None:
            # Evaluate the model function at equally-spaced points.
            # To determine the domain of x, use xlim if available. Otherwise,
            # use the range of x points measured up to this point.
            if self._xlim is None:
                x_data = self.livefit.independent_vars_data[self.__x_key]
                xmin, xmax = np.min(x_data), np.max(x_data)
            else:
                xmin, xmax = self._xlim
            x_points = np.linspace(xmin, xmax, self.num_points)
            kwargs = {self.__x_key: x_points}
            kwargs.update(self.livefit.result.values)
            self.y_data = self.livefit.result.model.eval(**kwargs)
            self.x_data = x_points
            # update kwargs to initial guess
            kwargs.update(self.livefit.result.init_values)
            self.update_plot()
        # Intentionally override LivePlot.event. Do not call super().

    def clear_plot(self):
        """Empties the data of the current plot line."""
        if self.current_line:
            self.current_line.set_data([], [])

    def update_plot(self):
        """Sets the current x and y data, then calls the `parent_plot` to update."""
        self.current_line.set_data(self.x_data, self.y_data)
        self.parent_plot.update_plot()
        if self.display_values:
            for text_object in self.text_objects:
                text_object.remove()
            self.text_objects.clear()
            vals = self.livefit.result.values
            if self.line_position is None:
                self.line_position = self.parent_plot.line_number
                self.parent_plot.line_number += len(vals)
            color = self.current_line.get_color()
            for i, (key, value) in enumerate(vals.items()):
                text_object = self.parent_plot.ax.text(
                    0.05,
                    0.95 - (i + self.line_position) * 0.05,
                    f"{key}: {value:.2e}",
                    transform=self.ax.transAxes,
                    verticalalignment="top",
                    color=color,
                )
                self.text_objects.append(text_object)


class Plot_Options(Ui_Plot_Options, QWidget):
    """Widget for setting the options of a plot.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget.
    ax : matplotlib.axes.Axes, optional
        The axes to plot on.
    livePlot : LivePlot, optional
        The LivePlot to set the options for.
    ax2 : matplotlib.axes.Axes, optional
        The second (y) axes to plot on.
    """

    def __init__(self, parent=None, ax=None, livePlot=None, ax2=None):
        super().__init__(parent)
        self.setupUi(self)
        self.ax = ax
        self.ax2 = ax2
        self.livePlot = livePlot
        self.livePlot.setup_done.connect(self.setup_table)
        self.checkBox_log_x.clicked.connect(self.set_log)
        self.checkBox_log_y.clicked.connect(self.set_log)
        self.checkBox_log_y2.clicked.connect(self.set_log)
        self.checkBox_use_abs_x.clicked.connect(self.set_log)
        self.checkBox_use_abs_y.clicked.connect(self.set_log)
        self.checkBox_use_abs_y2.clicked.connect(self.set_log)
        self.color_widges = []
        self.marker_widges = []
        self.linestyle_widges = []
        self.axis_widges = []
        self.marker_dict = {}
        self.linestyle_dict = {}

    def setup_table(self, x=None, y=None):
        """Sets up the table with the current lines in the plot."""
        self.tableWidget.setColumnCount(5)
        self.tableWidget.setMinimumWidth(400)
        self.tableWidget.setHorizontalHeaderLabels(
            ["Name", "marker", "color", "linestyle", "y-axis"]
        )
        self.tableWidget.verticalHeader().setHidden(True)
        for i, line in enumerate(self.ax.lines):
            if not self.marker_dict:
                self.marker_dict = {v: k for k, v in line.markers.items()}
                self.linestyle_dict = {
                    "".join(v.split("_")[2:]): k for k, v in line.lineStyles.items()
                }
            self.tableWidget.insertRow(i)
            item = QTableWidgetItem(line.get_label())
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.tableWidget.setItem(i, 0, item)

            markerwidge = QComboBox()
            markerwidge.addItems(self.marker_dict.keys())
            markerwidge.setCurrentText(line.markers[line.get_marker()])
            markerwidge.currentTextChanged.connect(lambda x, n=i: self.change_marker(n))
            self.marker_widges.append(markerwidge)
            self.tableWidget.setCellWidget(i, 1, markerwidge)

            colorwidge = QPushButton(line.get_color())
            colorwidge.clicked.connect(lambda n=i: self.change_color(n))
            self.color_widges.append(colorwidge)
            self.tableWidget.setCellWidget(i, 2, colorwidge)

            linestylewidge = QComboBox()
            linestylewidge.addItems(self.linestyle_dict.keys())
            # print(line.get_linestyle)
            linestylewidge.setCurrentText(
                "".join(line.lineStyles[line.get_linestyle()].split("_")[2:])
            )
            linestylewidge.currentTextChanged.connect(
                lambda x, n=i: self.change_linestyle(n)
            )
            self.linestyle_widges.append(linestylewidge)
            self.tableWidget.setCellWidget(i, 3, linestylewidge)

            # axiswidge = QComboBox()
            # axiswidge.addItems(['1', '2'])
            # axiswidge.setCurrentText('1')
            # axiswidge.currentTextChanged.connect(lambda x, n=i: self.change_axis(n))
            # self.axis_widges.append(axiswidge)
            # self.tableWidget.setCellWidget(i, 4, axiswidge)
        self.tableWidget.resizeColumnsToContents()

    def change_linestyle(self, row):
        """Changes the linestyle of the selected line in the plot.

        Parameters
        ----------
        row : int
            The row of the line in the table.
        """
        linestyle = self.linestyle_widges[row].currentText()
        name = self.ax.lines[row].get_label()
        self.livePlot.current_lines[name].set_linestyle(self.linestyle_dict[linestyle])
        self.ax.lines[row].set_linestyle(self.linestyle_dict[linestyle])
        self.ax.figure.canvas.draw_idle()

    def change_marker(self, row):
        """

        Parameters
        ----------
        row :


        Returns
        -------

        """
        marker = self.marker_widges[row].currentText()
        name = self.ax.lines[row].get_label()
        self.livePlot.current_lines[name].set_marker(self.marker_dict[marker])
        self.ax.lines[row].set_marker(self.marker_dict[marker])
        self.ax.figure.canvas.draw_idle()

    def change_color(self, row):
        """

        Parameters
        ----------
        row :


        Returns
        -------

        """
        color = QColorDialog.getColor()
        if color.isValid():
            self.color_widges[row].setText(color.name())
            line = self.ax.lines[row]
            line.set_color(color.name())
            name = self.ax.lines[row].get_label()
            self.livePlot.current_lines[name].set_color(color.name())
            self.ax.figure.canvas.draw_idle()

    def set_log(self):
        """ """
        x = self.checkBox_log_x.isChecked()
        x_scale = "log" if x else "linear"
        self.checkBox_use_abs_x.setEnabled(x)
        y = self.checkBox_log_y.isChecked()
        y_scale = "log" if y else "linear"
        self.checkBox_use_abs_y.setEnabled(y)
        y2 = self.checkBox_log_y2.isChecked()
        y2_scale = "log" if y2 else "linear"
        self.checkBox_use_abs_y2.setEnabled(y2)
        self.ax.set_xscale(x_scale)
        self.ax.set_yscale(y_scale)
        if self.ax2:
            self.ax2.set_yscale(y2_scale)
        self.livePlot.use_abs = {
            "x": self.checkBox_use_abs_x.isChecked() and x,
            "y": self.checkBox_use_abs_y.isChecked() and y,
            "y2": self.checkBox_use_abs_y2.isChecked() and y2,
        }
        self.livePlot.update_plot()


class MultiLivePlot(LivePlot, QObject):
    """ """

    new_data = Signal()
    setup_done = Signal()

    def __init__(
        self,
        ys=(),
        x=None,
        *,
        legend_keys=None,
        xlim=None,
        ylim=None,
        ax=None,
        epoch="run",
        xlabel="",
        ylabel="",
        evaluator=None,
        title="",
        stream_name="primary",
        do_plot=True,
        fitPlots=None,
        multi_stream=False,
        y_axes=None,
        ax2=None,
        ylabel2="",
        **kwargs,
    ):
        LivePlot.__init__(
            self,
            y=ys[0],
            x=x,
            legend_keys=legend_keys,
            xlim=xlim,
            ylim=ylim,
            ax=ax,
            epoch=epoch,
            **kwargs,
        )
        QObject.__init__(self)
        self.use_abs = {"x": False, "y": False, "y2": False}
        self.__setup_lock = threading.Lock()
        self.__setup_event = threading.Event()
        self.eva = evaluator
        self.stream_name = stream_name
        self.desc = []
        self.do_plot = do_plot
        self.multi_stream = multi_stream
        self.fitPlots = fitPlots or []
        self.y_axes = y_axes or {}
        self.ax2 = ax2
        for fit in self.fitPlots:
            fit.parent_plot = self
        if isinstance(ys, str):
            ys = [ys]

        def setup():
            """ """
            # Run this code in start() so that it runs on the correct thread.
            nonlocal ys, x, legend_keys, xlim, ylim, ax, epoch, kwargs, xlabel, ylabel, title
            with self.__setup_lock:
                if self.__setup_event.is_set():
                    return
                self.__setup_event.set()
            if ax is None:
                fig, ax = plt.subplots()
            self.ax = ax

            if legend_keys is None:
                legend_keys = ys
            self.legend_keys = legend_keys
            if x is not None:
                self.x, *others = get_obj_fields([x])
            else:
                self.x = "seq_num"
            self.ys = get_obj_fields(ys)
            # a = ylabel or ys[0]
            # print(a)
            self.ax.set_ylabel(ylabel or ys[0])
            self.ax.set_xlabel(xlabel or x or "sequence #")
            if title:
                self.ax.set_title(title)
            if xlim is not None:
                self.ax.set_xlim(*xlim)
            if ylim is not None:
                self.ax.set_ylim(*ylim)
            self.ax.margins(0.1)
            self.kwargs = kwargs
            self.lines = []
            self.legend = None
            # self.legend_title = " :: ".join([name for name in self.legend_keys])
            self._epoch_offset = None  # used if x == 'time'
            self._epoch = epoch

        self.xlabel = xlabel
        self.ylabel = ylabel
        self.ylabel2 = ylabel2
        self.__setup = setup
        self._epoch_offset = 0
        self.x_data = []
        self.y_data = {}
        self.maxlen = np.inf
        self.y_names = ys
        self.ys = get_obj_fields(ys)
        for y in ys:
            self.y_data[y] = []
        self.current_lines = {}
        self.descs_fit_readying = {}
        self.legend = None
        self.line_number = 0

    def change_maxlen(self, maxlen):
        """

        Parameters
        ----------
        maxlen :


        Returns
        -------

        """
        self.maxlen = maxlen
        if maxlen < np.inf:
            self.x_data = deque(self.x_data, maxlen=maxlen)
            for y in self.y_data:
                self.y_data[y] = deque(self.y_data[y], maxlen=maxlen)
        else:
            self.x_data = list(self.x_data)
            for y in self.y_data:
                self.y_data[y] = list(self.y_data[y])

    def start(self, doc):
        """

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        self.__setup()
        # The doc is not used; we just use the signal that a new run began.
        # self._epoch_offset = doc['time']  # used if self.x == 'time'
        # self.x_data = []
        # self.y_data = {}
        # for y in self.ys:
        #     self.y_data[y] = []
        # label = " :: ".join(
        # [str(doc.get(name, name)) for name in self.legend_keys])
        kwargs = ChainMap({"ls": "None", "marker": "x"})
        self.current_lines = {}
        for i, y in enumerate(self.ys):
            try:
                if y in self.y_axes and self.y_axes[y] == 2:
                    (self.current_lines[y],) = self.ax2.plot(
                        [], [], label=self.legend_keys[i], color=stdCols[i], **kwargs
                    )
                    self.ax.plot(
                        [], [], label=self.legend_keys[i], color=stdCols[i], **kwargs
                    )
                else:
                    (self.current_lines[y],) = self.ax.plot(
                        [], [], label=self.legend_keys[i], color=stdCols[i], **kwargs
                    )
            except Exception as e:
                print(e)
        self.lines.append(self.current_lines)
        legend = self.ax.legend(loc=0)
        try:
            # matplotlib v3.x
            self.legend = legend.set_draggable(True)
        except AttributeError:
            # matplotlib v2.x (warns in 3.x)
            self.legend = legend.draggable(True)
        for fit in self.fitPlots:
            fit.start(doc)
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        if self.ax2:
            self.ax2.set_ylabel(self.ylabel2)
        self.setup_done.emit()

    def descriptor(self, doc):
        """

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        if doc["name"] == self.stream_name:
            self.desc.append(doc["uid"])
        elif doc["name"].startswith(f"{self.stream_name}_fits_readying_"):
            for fit in self.fitPlots:
                if (
                    doc["name"]
                    == f"{self.stream_name}_fits_readying_{fit.livefit.name}"
                ):
                    self.descs_fit_readying[doc["uid"]] = fit
        elif self.multi_stream and doc["name"].startswith(self.stream_name):
            self.desc.append(doc["uid"])

    def clear_plot(self):
        """ """
        self.x_data.clear()
        for y in self.y_data:
            self.y_data[y].clear()

    def event(self, doc):
        """Unpack data from the event and call self.update_plot().

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        # This outer try/except block is needed because multiple event
        # streams will be emitted by the RunEngine and not all event
        # streams will have the keys we want.
        # This inner try/except block handles seq_num and time, which could
        # be keys in the data or accessing the standard entries in every
        # event.
        if doc["descriptor"] not in self.desc:
            if doc["descriptor"] in self.descs_fit_readying:
                self.descs_fit_readying[doc["descriptor"]].get_ready()
            return
        try:
            new_x = doc["data"][self.x]
        except KeyError:
            if self.x in ("time", "seq_num"):
                new_x = doc[self.x]
            else:
                if not self.eva.is_to_date(doc["time"]):
                    self.eva.event(doc)
                new_x = self.eva.eval(self.x)

        new_y = {}
        for y in self.ys:
            try:
                new_y[y] = doc["data"][y]
            except KeyError:
                if not self.eva.is_to_date(doc["time"]):
                    self.eva.event(doc)
                new_y[y] = self.eva.eval(y)

        # Special-case 'time' to plot against against experiment epoch, not
        # UNIX epoch.
        if self.x == "time" and self._epoch == "run":
            new_x -= self._epoch_offset

        self.update_caches(new_x, new_y)
        for fit in self.fitPlots:
            fit.event(doc)
        self.update_plot()
        # super().event(doc)

    def update_caches(self, x, ys):
        """

        Parameters
        ----------
        x :

        ys :


        Returns
        -------

        """
        for y in ys:
            self.y_data[y].append(ys[y])
        self.x_data.append(x)

    def update_plot(self):
        """ """
        for y, line in self.current_lines.items():
            xdat = np.abs(self.x_data) if self.use_abs["x"] else self.x_data
            ydat = np.abs(self.y_data[y]) if self.use_abs["y"] else self.y_data[y]
            line.set_data(xdat, ydat)
        # Rescale and redraw.
        try:
            self.ax.relim(visible_only=True)
            self.ax.autoscale_view(tight=True)
            if self.ax2:
                self.ax2.relim(visible_only=True)
                self.ax2.autoscale_view(tight=True)
            self.ax.figure.canvas.draw_idle()
            self.new_data.emit(None)
        except:
            pass

    def stop(self, doc):
        """

        Parameters
        ----------
        doc :


        Returns
        -------

        """
        if not self.x_data:
            print(
                "MultiLivePlot did not get any data that corresponds to the "
                "x axis. {}".format(self.x)
            )
        for y in self.y_data:
            if not self.y_data[y]:
                print(
                    "MultiLivePlot did not get any data that corresponds to the "
                    "y axis. {}".format(y)
                )
            if len(self.y_data[y]) != len(self.x_data):
                print(
                    "MultiLivePlot has a different number of elements for x ({}) and"
                    "y ({}, {})".format(len(self.x_data), len(self.y_data), y)
                )
        for fit in self.fitPlots:
            fit.stop(doc)


class PlotWidget_NoBluesky(QWidget):
    """ """

    def __init__(
        self,
        xlabel="",
        ylabel="",
        parent=None,
        title="",
        ylabel2="",
        y_axes=None,
        labels=(),
        first_hidden=None,
        show_plot=True,
        maxlen=np.inf,
    ):
        app = QCoreApplication.instance()
        if app is None:
            app = QApplication(sys.argv)
        super().__init__(parent=parent)
        canvas = MPLwidget()
        self.ax = canvas.axes
        self.plot = MultiPlot_NoBluesky(
            self.ax, xlabel, ylabel, ylabel2, y_axes, labels, first_hidden, show_plot
        )
        self.toolbar = NavigationToolbar2QT(canvas, self)

        self.pushButton_show_options = QPushButton("Show Options")
        self.pushButton_show_options.clicked.connect(self.show_options)
        self.pushButton_autoscale = QPushButton("Autoscale")
        self.pushButton_autoscale.clicked.connect(self.autoscale)
        self.pushButton_clear = QPushButton("Clear Plot")
        self.pushButton_clear.clicked.connect(self.clear_plot)
        self.plot_options = Plot_Options(self, self.ax, self.plot)
        label_n_data = QLabel("# data points:")
        self.lineEdit_n_data = QLineEdit(str(maxlen))
        self.lineEdit_n_data.returnPressed.connect(self.change_maxlen)
        self.change_maxlen()

        self.setWindowTitle(title or f"{xlabel} vs. {ylabel}")
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))

        layout = QGridLayout()
        layout.addWidget(canvas, 0, 1, 1, 6)
        layout.addWidget(self.toolbar, 1, 4)
        layout.addWidget(self.pushButton_show_options, 1, 1)
        layout.addWidget(self.pushButton_autoscale, 1, 2)
        layout.addWidget(self.pushButton_clear, 1, 3)
        layout.addWidget(label_n_data, 1, 5)
        layout.addWidget(self.lineEdit_n_data, 1, 6)
        layout.addWidget(self.plot_options, 0, 0, 2, 1)
        self.setLayout(layout)

        self.plot_options.hide()
        self.options_open = False
        self.show()

    def change_maxlen(self):
        """ """
        text = self.lineEdit_n_data.text()
        if not text:
            maxlen = np.inf
        else:
            try:
                maxlen = int(text)
            except:
                return
        self.plot.change_maxlen(maxlen)

    def clear_plot(self):
        """Clear the plot by removing the data from the plot and clearing all
        fit plots.

        Parameters
        ----------

        Returns
        -------

        """
        self.plot.clear_plot()

    def autoscale(self):
        """ """
        self.ax.autoscale()
        self.plot.ax2.autoscale()
        self.ax.figure.canvas.draw_idle()

    def show_options(self):
        """ """
        if self.options_open:
            self.pushButton_show_options.setText("Show Options")
            self.options_open = False
            self.plot_options.hide()
        else:
            self.pushButton_show_options.setText("Hide Options")
            self.options_open = True
            self.plot_options.show()
        self.adjustSize()


class MultiPlot_NoBluesky(QObject):
    """ """

    new_data = Signal()
    setup_done = Signal()

    def __init__(
        self,
        ax,
        xlabel="",
        ylabel="",
        ylabel2="",
        y_axes=None,
        labels=(),
        first_hidden=None,
        show_plot=True,
    ):
        super().__init__()
        self.ax = ax
        self.ax.set_xlabel(xlabel)
        self.ax.set_ylabel(ylabel)
        self.ax2 = self.ax.twinx()
        self.ax2.set_ylabel(ylabel2)
        self.labels = labels
        self.maxlen = np.inf
        self.xdata = []
        self.ydata = {}
        self.current_lines = {}
        self.y_axes = y_axes or {}
        self.first_hidden = first_hidden or []
        self.show_plot = show_plot
        self.use_abs = {"x": False, "y": False, "y2": False}

    def add_data(self, x, ys, add=True):
        """

        Parameters
        ----------
        x :

        ys :

        add :
             (Default value = True)

        Returns
        -------

        """
        if not self.current_lines:
            if len(self.labels) != len(ys):
                self.labels = list(ys.keys())
            for i, y in enumerate(ys):
                try:
                    if y in self.y_axes and self.y_axes[y] == 2:
                        (self.current_lines[y],) = self.ax2.plot(
                            [],
                            [],
                            linestyle="" if y in self.first_hidden else "-",
                            label=self.labels[i],
                            color=stdCols[i],
                        )
                        self.ax.plot(
                            [],
                            [],
                            linestyle="" if y in self.first_hidden else "-",
                            label=self.labels[i],
                            color=stdCols[i],
                        )
                    else:
                        (self.current_lines[y],) = self.ax.plot(
                            [],
                            [],
                            linestyle="" if y in self.first_hidden else "-",
                            label=self.labels[i],
                            color=stdCols[i],
                        )
                    if self.maxlen < np.inf:
                        self.ydata[y] = deque(maxlen=self.maxlen)
                    else:
                        self.ydata[y] = []
                except Exception as e:
                    print(e)
            self.setup_done.emit()
        if add:
            self.xdata.append(x)
            for y in ys:
                if y in self.ydata:
                    self.ydata[y].append(ys[y])
        else:
            self.xdata = [x]
            self.ydata.clear()
            for y in ys:
                self.ydata[y] = [ys[y]]
        if self.show_plot:
            self.update_plot()

    def change_maxlen(self, maxlen):
        """

        Parameters
        ----------
        maxlen :


        Returns
        -------

        """
        self.maxlen = maxlen
        if maxlen < np.inf:
            self.xdata = deque(self.xdata, maxlen=maxlen)
            for y in self.ydata:
                self.ydata[y] = deque(self.ydata[y], maxlen=maxlen)
        else:
            self.xdata = list(self.xdata)
            for y in self.ydata:
                self.ydata[y] = list(self.ydata[y])

    def update_plot(self):
        """ """
        for y, line in self.current_lines.items():
            xdat = np.abs(self.xdata) if self.use_abs["x"] else self.xdata
            ydat = np.abs(self.ydata[y]) if self.use_abs["y"] else self.ydata[y]
            line.set_data(xdat, ydat)
        # Rescale and redraw.
        self.ax.relim(visible_only=True)
        self.ax.autoscale_view(tight=True)
        self.ax.legend()
        self.ax.figure.canvas.draw_idle()
        self.new_data.emit()

    def clear_plot(self):
        """ """
        self.xdata.clear()
        for y in self.ydata:
            self.ydata[y].clear()


# if __name__ == '__main__':
#     from bluesky import RunEngine
#     from bluesky.plans import scan
#     from ophyd.sim import motor, det
#
#     motor.delay = 0.1
#
#     def plan():
#         for i in range(4, 5):
#             yield from scan([det], motor, -5, 5, 3**i)
#
#     RE = RunEngine()
#     app = QApplication(sys.argv)
#     # myapp = PlotWidget(run_engine=RE, x_name='motor', y_names=['det'], title='test', xlabel='aaaa', ylabel='bbbb')
#     myapp = PlotWidget('motor', ['det', 'det**2', 'sin(motor)'])
#     myapp.show()
#     RE.subscribe(myapp.livePlot)
#     RE(plan())
#     sys.exit(app.exec())



from PySide6.QtWidgets import QTextEdit, QWidget
from PySide6.QtCore import Signal
from PySide6.QtGui import QTextCursor
from nomad_camels.utility import variables_handling

import io

import sys


class Error_Writer(io.StringIO):
    """ """

    def __init__(self, text_signal_handler, use_old_stderr=False):
        super().__init__()
        self.text_signal_handler = text_signal_handler
        self.old_stderr = sys.stderr
        self.use_old_stderr = use_old_stderr

    def write(self, *args, **kwargs):
        """

        Parameters
        ----------
        *args :

        **kwargs :


        Returns
        -------

        """
        text = args[0]
        if self.use_old_stderr:
            self.old_stderr.write(text)
        self.text_signal_handler.write_error_signal.emit(text)


class Text_Writer(io.StringIO):
    """ """

    def __init__(self, text_signal_handler, use_old_stdout=False):
        super().__init__()
        self.text_signal_handler = text_signal_handler
        self.old_stdout = sys.stdout
        self.use_old_stdout = use_old_stdout

    def write(self, *args, **kwargs):
        """

        Parameters
        ----------
        *args :

        **kwargs :


        Returns
        -------

        """
        text = args[0]
        if self.use_old_stdout:
            self.old_stdout.write(text)
        if text == "\n":
            return
        self.text_signal_handler.write_output_signal.emit(text)


class TextSignalHanlder(QWidget):
    write_error_signal = Signal(str)
    write_output_signal = Signal(str)


class Console_TextEdit(QTextEdit):
    """ """

    def __init__(self, parent, max_lines=10000):
        super().__init__(parent)
        use_old = not sys.executable.endswith("pythonw.exe")
        self.text_handler = TextSignalHanlder()
        self.text_handler.write_output_signal.connect(self.write_output)
        self.text_handler.write_error_signal.connect(self.write_error)
        self.text_writer = Text_Writer(self.text_handler, use_old)
        self.error_writer = Error_Writer(self.text_handler, use_old)
        self.max_lines = max_lines

    def write_error(self, text):
        self.setTextColor(variables_handling.get_color("strong_red"))
        self.append(text)

    def write_output(self, text):
        self.setTextColor(variables_handling.get_color("black"))
        self.append(text)

    def append(self, text):
        """Ensure the number of lines does not exceed `self.max_lines`."""
        super().append(text)
        document = self.document()
        line_count = document.blockCount()

        # Remove excess lines from the beginning
        if line_count > self.max_lines:
            cursor = QTextCursor(document)
            cursor.movePosition(
                QTextCursor.Start
            )  # Move cursor to the start of the document
            for _ in range(line_count - self.max_lines):
                cursor.select(QTextCursor.BlockUnderCursor)  # Select the first block
                cursor.removeSelectedText()  # Remove the selected text
                cursor.deleteChar()  # Remove the newline character



import numpy as np
from PySide6.QtWidgets import QWidget, QTableWidgetItem, QLabel, QGridLayout, QLineEdit
from PySide6.QtCore import Qt

from nomad_camels.main_classes.loop_step import Loop_Step_Container, Loop_Step_Config
from nomad_camels.utility.number_formatting import format_number
from nomad_camels.utility import variables_handling
from nomad_camels.ui_widgets.variable_tool_tip_box import Variable_Box

from nomad_camels.gui.for_loop import Ui_for_loop_config


class While_Loop_Step(Loop_Step_Container):
    """A loopstep that adds a simple While Loop with a condition, that
    may be just written as python-code.

    Attributes
    ----------
    condition : str
        The condition which is used for the loop. It should be interpretable
        python-code within the namespace of the protocol.
    expected_iterations : int, default 1
        The expected number of iterations, used for the steps time weight for
        the progress bar. Ideally rather too large than to small (otherwise the
        progress bar reaches 100% too early).
    """

    def __init__(
        self, name="", children=None, parent_step=None, step_info=None, **kwargs
    ):
        super().__init__(name, children, parent_step, step_info, **kwargs)
        self.step_type = "While Loop"
        if step_info is None:
            step_info = {}
        self.condition = step_info["condition"] if "condition" in step_info else "!"
        self.expected_interations = (
            step_info["expected_interations"]
            if "expected_interations" in step_info
            else 1
        )

    def update_variables(self):
        """Here the Count of the while-loop is included as a variable."""
        variables = {f'{self.name.replace(" ", "_")}_Count': 0}
        for variable in variables:
            if variable in variables_handling.loop_step_variables:
                raise Exception("Variable already defined!")
        variables_handling.loop_step_variables.update(variables)
        super().update_variables()

    def get_protocol_string(self, n_tabs=1):
        """The string consists of declaring the count-variable. Then the
        while loop with the desired condition is started. After all the
        children-steps, the count-variable increased by 1."""
        self.update_time_weight()
        tabs = "\t" * n_tabs
        count_var = f'{self.name.replace(" ", "_")}_Count'
        protocol_string = super().get_protocol_string(n_tabs)
        protocol_string += f"{tabs}{count_var} = 0\n"
        protocol_string += f'{tabs}namespace["{count_var}"] = {count_var}\n'
        protocol_string += f'{tabs}while eva.eval("{self.condition}"):\n'
        protocol_string += f'{tabs}\tnamespace["{count_var}"] = {count_var}\n'
        protocol_string += self.get_children_strings(n_tabs + 1)
        protocol_string += f"{tabs}\t{count_var} += 1\n"
        return protocol_string

    def update_time_weight(self):
        """Multiplies the time_weight of the children with the expected
        iterations + 5"""
        super().update_time_weight()
        self.time_weight *= self.expected_interations + 5

    def get_protocol_short_string(self, n_tabs=0):
        """Adds the condition to the string"""
        tabs = "\t" * n_tabs
        short_string = f"{tabs}'{self.name}' while {self.condition}:\n"
        for child in self.children:
            short_string += child.get_protocol_short_string(n_tabs + 1)
        return short_string


class While_Loop_Step_Config(Loop_Step_Config):
    """Configuration-Widget for the while-loop step."""

    def __init__(self, loop_step: While_Loop_Step, parent=None):
        super().__init__(parent, loop_step)
        self.sub_widget = While_Loop_Step_Config_Sub(loop_step, self)
        self.layout().addWidget(self.sub_widget, 1, 0, 1, 5)


class While_Loop_Step_Config_Sub(QWidget):
    """Sub-config for the while-loop. It consists only of a single line
    that takes the condition for the while loop.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, loop_step: While_Loop_Step, parent=None):
        super().__init__(parent)
        self.loop_step = loop_step

        label = QLabel("Condition:")
        label_it = QLabel("Expected number of iterations:")
        self.lineEdit_condition = Variable_Box(self)
        self.lineEdit_condition.setText(loop_step.condition)
        self.lineEdit_expection = QLineEdit(str(loop_step.expected_interations))
        self.lineEdit_condition.textChanged.connect(self.update_condition)
        self.lineEdit_expection.textChanged.connect(self.update_condition)

        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(label, 0, 0)
        layout.addWidget(self.lineEdit_condition, 0, 1)
        layout.addWidget(label_it, 1, 0)
        layout.addWidget(self.lineEdit_expection, 1, 1)
        self.setLayout(layout)

    def update_condition(self):
        """Saves the condition into the loop_step."""
        self.loop_step.condition = self.lineEdit_condition.text()
        try:
            self.loop_step.expected_interations = float(self.lineEdit_expection.text())
        except:
            pass


class For_Loop_Step(Loop_Step_Container):
    """Loop_Step representing a For Loop. It offers several ways of
    defining the sweep.

    Attributes
    ----------
    loop_type : str
        can be one of the following:
        - "start - stop" the loop goes from `min_val` to `max_val`
        - "start - min - max - stop" or "start - max - min - stop" goes from start over min/max, max/min to stop
        - "Value-List" the loop uses the values inside `val_list`
        - "Text-File" the loop uses the values of a single-column text file given by `file_path`
    sweep_mode : str
        only relevant if not using "Value-List" or "Text-File" mode
        can be:
        - "linear" the loop goes over a simple linear space
        - "logarithmic" the steps increase logarithmically
        - "exponential" the steps increase exponentially
        - "1/x" the steps increase with 1/x
    start_val : float
        the first value for the loop
    stop_val : float
        the last value for the loop
    min_val : float
        the minimum value for the loop
    max_val : float
        the maximum value for the loop
    n_points : int
        the number of points the loop iterates over. If using the max
        and min-vals, the distance between those is used, thus in total
        creating more steps
    n_iterations : int
        number of runs the loop will do
    val_list : list of float
        a list of all the steps, the loop iterates over, used if
        "Value-List" is chosen
    file_path : str or path
        the path where the file for "Text-File" is lying
    include_end_points : bool
        whether to include the `stop_val` into the iteration
    """

    def __init__(
        self, name="", children=None, parent_step=None, step_info=None, **kwargs
    ):
        super().__init__(name, children, parent_step, step_info, **kwargs)
        self.step_type = "For Loop"
        if step_info is None:
            step_info = {}
        self.loop_type = (
            step_info["loop_type"] if "loop_type" in step_info else "start - stop"
        )
        self.start_val = step_info["start_val"] if "start_val" in step_info else ""
        self.stop_val = step_info["stop_val"] if "stop_val" in step_info else ""
        self.min_val = step_info["min_val"] if "min_val" in step_info else ""
        self.max_val = step_info["max_val"] if "max_val" in step_info else ""
        self.n_points = step_info["n_points"] if "n_points" in step_info else ""
        self.sweep_mode = (
            step_info["sweep_mode"] if "sweep_mode" in step_info else "linear"
        )
        self.n_iterations = (
            step_info["n_iterations"] if "n_iterations" in step_info else 0
        )
        # self.point_array = step_info['point_array'] if 'point_array' in step_info else []
        self.val_list = step_info["val_list"] if "val_list" in step_info else []
        self.file_path = step_info["file_path"] if "file_path" in step_info else ""
        self.include_end_points = (
            step_info["include_end_points"]
            if "include_end_points" in step_info
            else True
        )
        self.use_distance = (
            step_info["use_distance"] if "use_distance" in step_info else False
        )
        self.point_distance = (
            step_info["point_distance"] if "point_distance" in step_info else ""
        )
        # self.update_variables()

    def update_variables(self):
        """Includes the value and iteration-count of the loop."""
        variables = {
            f'{self.name.replace(" ", "_")}_Count': 0,
            f'{self.name.replace(" ", "_")}_Value': 0,
        }
        for variable in variables:
            if variable in variables_handling.loop_step_variables:
                raise Exception("Variable already defined!")
        variables_handling.loop_step_variables.update(variables)
        super().update_variables()

    def get_protocol_string(self, n_tabs=1):
        """The loop is enumerating over the selected points. When using a range,
        the get_range helper function is used."""
        tabs = "\t" * n_tabs
        if self.loop_type in [
            "start - stop",
            "start - min - max - stop",
            "start - max - min - stop",
        ]:
            enumerator = f'helper_functions.get_range(eva, "{self.start_val}", "{self.stop_val}", "{self.n_points}", "{self.min_val or np.nan}", "{self.max_val or np.nan}", "{self.loop_type}", "{self.sweep_mode}", "{self.include_end_points}", "{self.point_distance}", {self.use_distance})'
            # enumerator = get_space_string(self.start_val, self.stop_val,
            #                               self.n_points, self.min_val,
            #                               self.max_val, self.loop_type,
            #                               self.sweep_mode,
            #                               self.include_end_points)
        elif self.loop_type == "Value-List":
            enumerator = self.val_list
        else:
            enumerator = f'np.loadtxt("{self.file_path}")'
        protocol_string = super().get_protocol_string(n_tabs)
        protocol_string += f'{tabs}for {self.name.replace(" ", "_")}_Count, {self.name.replace(" ", "_")}_Value in enumerate({enumerator}):\n'
        protocol_string += f'{tabs}\tnamespace.update({{"{self.name.replace(" ", "_")}_Count": {self.name.replace(" ", "_")}_Count, "{self.name.replace(" ", "_")}_Value": {self.name.replace(" ", "_")}_Value}})\n'
        protocol_string += self.get_children_strings(n_tabs + 1)
        self.update_time_weight()
        return protocol_string

    def update_time_weight(self):
        """Multiplies the children time_weight (-1 for the step itself) by the
        number of iterations and adds 1 (for the step itself)."""
        super().update_time_weight()
        self.time_weight = (self.time_weight - 1) * self.n_iterations + 1

    def get_protocol_short_string(self, n_tabs=0):
        """Shows the type of loop and the range in the short string as well"""
        tabs = "\t" * n_tabs
        if self.loop_type == "Value-List":
            vals = self.val_list
        elif self.loop_type in [
            "start - stop",
            "start - min - max - stop",
            "start - max - min - stop",
        ]:
            vals = f"(start: {self.start_val}, stop: {self.stop_val}, "
            if "max" in self.loop_type:
                vals += f"min: {self.min_val}, max: {self.max_val}, "
            vals += f"points: {self.n_points})"
        else:
            vals = self.file_path
        short_string = f"{tabs}'{self.name}' for {vals}:\n"
        for child in self.children:
            short_string += child.get_protocol_short_string(n_tabs + 1)
        return short_string


class For_Loop_Step_Config(Loop_Step_Config):
    """Configuration-Widget for the for-loop step."""

    def __init__(self, loop_step: For_Loop_Step, parent=None):
        super().__init__(parent, loop_step)
        self.sub_widget = For_Loop_Step_Config_Sub(parent=self, loop_step=loop_step)
        self.layout().addWidget(self.sub_widget, 1, 0, 1, 5)


class For_Loop_Step_Config_Sub(Ui_for_loop_config, QWidget):
    """Provides the main config for the For Loop."""

    def __init__(self, loop_step: For_Loop_Step, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.loop_step = loop_step
        self.load_data()
        self.loop_type_change()
        self.build_preview_array()

        self.comboBox_loop_type.currentTextChanged.connect(self.loop_type_change)
        self.lineEdit_start.textChanged.connect(self.change_dist_n_mode)
        self.lineEdit_stop.textChanged.connect(self.change_dist_n_mode)
        self.lineEdit_min.textChanged.connect(self.change_dist_n_mode)
        self.lineEdit_max.textChanged.connect(self.change_dist_n_mode)
        self.lineEdit_point_distance.textChanged.connect(self.change_point_dist)
        self.lineEdit_n_points.textChanged.connect(self.change_n_points)
        self.checkBox_include_endpoints.clicked.connect(self.change_dist_n_mode)
        self.comboBox_sweep_mode.currentTextChanged.connect(self.change_sweep_mode)
        self.pushButton_del_point.clicked.connect(self.del_point)
        self.pushButton_add_point.clicked.connect(self.add_point)
        self.radioButton_n_points.clicked.connect(self.change_dist_n_mode)
        self.radioButton_point_distance.clicked.connect(self.change_dist_n_mode)

        self.tableWidget_points.cellChanged.connect(self.value_list_changed)
        self.path_line_button.path_changed.connect(self.build_preview_array)

        self.building = False
        self.change_dist_n_mode()

    def load_data(self):
        """Loads the data from the loop_step into the UI-widgets."""
        self.comboBox_loop_type.setCurrentText(self.loop_step.loop_type)
        self.lineEdit_start.setText(self.loop_step.start_val)
        self.lineEdit_stop.setText(self.loop_step.stop_val)
        self.lineEdit_min.setText(self.loop_step.min_val)
        self.lineEdit_max.setText(self.loop_step.max_val)
        self.lineEdit_n_points.setText(self.loop_step.n_points)
        self.lineEdit_point_distance.setText(self.loop_step.point_distance)
        self.radioButton_n_points.setChecked(not self.loop_step.use_distance)
        self.radioButton_point_distance.setChecked(self.loop_step.use_distance)
        self.comboBox_sweep_mode.setCurrentText(self.loop_step.sweep_mode)
        self.checkBox_include_endpoints.setChecked(self.loop_step.include_end_points)
        self.path_line_button.set_path(self.loop_step.file_path)

    def loop_type_change(self):
        """Enables / disables the respective elements that are used for
        the single loop-types, then builds the preview.

        Parameters
        ----------

        Returns
        -------

        """
        combo_text = self.comboBox_loop_type.currentText()
        if self.comboBox_loop_type.currentText() == "start - stop":
            self.path_line_button.setEnabled(False)
            self.sweep_widget.setEnabled(True)
            self.lineEdit_max.setEnabled(False)
            self.lineEdit_min.setEnabled(False)
            self.pushButton_add_point.setEnabled(False)
            self.pushButton_del_point.setEnabled(False)
            self.label_min.setEnabled(False)
            self.label_max.setEnabled(False)
        elif combo_text in ["start - min - max - stop", "start - max - min - stop"]:
            self.path_line_button.setEnabled(False)
            self.sweep_widget.setEnabled(True)
            self.lineEdit_max.setEnabled(True)
            self.lineEdit_min.setEnabled(True)
            self.pushButton_add_point.setEnabled(False)
            self.pushButton_del_point.setEnabled(False)
            self.label_min.setEnabled(True)
            self.label_max.setEnabled(True)
        elif combo_text == "Value-List":
            self.path_line_button.setEnabled(False)
            self.sweep_widget.setEnabled(False)
            self.pushButton_add_point.setEnabled(True)
            self.pushButton_del_point.setEnabled(True)
        else:
            self.path_line_button.setEnabled(True)
            self.sweep_widget.setEnabled(False)
            self.pushButton_add_point.setEnabled(False)
            self.pushButton_del_point.setEnabled(False)
        self.build_preview_array()
        self.loop_step.loop_type = combo_text

    def change_point_dist(self):
        """Updates the number of points when the distance between them
        is changed.

        Parameters
        ----------

        Returns
        -------

        """
        if self.radioButton_n_points.isChecked():
            return
        if self.building:
            return
        self.build_preview_array()
        points = self.loop_step.n_iterations
        self.building = True
        self.lineEdit_n_points.setText(str(points))
        self.building = False

    def change_n_points(self):
        """Updates the displayed distance between the points when their
        number is changed.

        Parameters
        ----------

        Returns
        -------

        """
        if self.radioButton_point_distance.isChecked():
            return
        if self.building:
            return
        self.build_preview_array()
        self.loop_step.include_end_points = self.checkBox_include_endpoints.isChecked()
        self.building = True
        self.lineEdit_point_distance.setText(format_number(self.distance))
        self.building = False

    def change_dist_n_mode(self):
        use_dist = self.radioButton_point_distance.isChecked()
        self.loop_step.use_distance = use_dist
        if use_dist:
            self.change_point_dist()
        else:
            self.change_n_points()
        self.lineEdit_n_points.setEnabled(not use_dist)
        self.lineEdit_point_distance.setEnabled(use_dist)

    def change_sweep_mode(self):
        """Enables / disables the point-distance widget corresponding to
        the selected sweep mode.

        Parameters
        ----------

        Returns
        -------

        """
        if self.comboBox_sweep_mode.currentText() == "linear":
            self.radioButton_point_distance.setEnabled(True)
        else:
            self.radioButton_point_distance.setEnabled(False)
            self.loop_step.use_distance = self.radioButton_point_distance.isChecked()
            self.radioButton_point_distance.setChecked(False)
            self.radioButton_n_points.setChecked(True)
        self.change_dist_n_mode()
        self.loop_step.sweep_mode = self.comboBox_sweep_mode.currentText()

    def build_preview_array(self):
        """Builds the displayed array to preview the steps, the for loop
        will make.

        Parameters
        ----------

        Returns
        -------

        """
        self.tableWidget_points.clear()
        self.tableWidget_points.setRowCount(0)
        if self.comboBox_loop_type.currentText() in [
            "start - stop",
            "start - min - max - stop",
            "start - max - min - stop",
        ]:
            try:
                start = self.lineEdit_start.text()
                self.loop_step.start_val = start
            except ValueError:
                start = np.nan
            try:
                stop = self.lineEdit_stop.text()
                self.loop_step.stop_val = stop
            except ValueError:
                stop = np.nan
            try:
                points = self.lineEdit_n_points.text()
                self.loop_step.n_points = points
            except ValueError:
                points = 0
            try:
                distance = self.lineEdit_point_distance.text()
                self.loop_step.point_distance = distance
            except ValueError:
                distance = np.nan
            use_distance = self.radioButton_point_distance.isChecked()
            self.loop_step.use_distance = use_distance
            try:
                min_val = self.lineEdit_min.text()
                self.loop_step.min_val = min_val
            except ValueError:
                min_val = np.nan
            try:
                max_val = self.lineEdit_max.text()
                self.loop_step.max_val = max_val
            except ValueError:
                max_val = np.nan
            from nomad_camels.bluesky_handling import (
                helper_functions,
                evaluation_helper,
            )

            min_val = min_val if min_val != "" else np.nan
            max_val = max_val if max_val != "" else np.nan
            start = start if start != "" else np.nan
            stop = stop if stop != "" else np.nan
            points = points if points != "" else 0
            distance = distance if distance != "" else np.nan
            namespace = {}
            namespace.update(variables_handling.protocol_variables)
            namespace.update(variables_handling.loop_step_variables)
            for channel in variables_handling.channels:
                namespace.update({channel: 1})
            try:
                vals = helper_functions.get_range(
                    evaluator=evaluation_helper.Evaluator(namespace=namespace),
                    start=start,
                    stop=stop,
                    points=points,
                    min_val=min_val,
                    max_val=max_val,
                    distance=distance,
                    loop_type=self.comboBox_loop_type.currentText(),
                    sweep_mode=self.comboBox_sweep_mode.currentText(),
                    endpoint=self.checkBox_include_endpoints.isChecked(),
                    use_distance=use_distance,
                )
            except (ValueError, ZeroDivisionError) as e:
                if e.args[0] == "Start and Stop must be between min and max!":
                    vals = "Start and Stop must be between Min and Max!".split()
                else:
                    return [np.nan]
        elif self.comboBox_loop_type.currentText() == "Value-List":
            vals = self.loop_step.val_list
        else:
            try:
                file = self.path_line_button.get_path()
                vals = np.loadtxt(file)
                self.loop_step.file_path = file
            except OSError:
                return
        self.tableWidget_points.setRowCount(len(vals) if len(vals) < 1001 else 1001)
        self.tableWidget_points.horizontalHeader().hide()
        self.tableWidget_points.setColumnCount(1)
        self.loop_step.n_iterations = len(vals)
        try:
            self.distance = vals[1] - vals[0] if len(vals) > 1 else np.nan
        except:
            self.distance = np.nan
        for i, val in enumerate(vals):
            item = QTableWidgetItem(val if isinstance(val, str) else format_number(val))
            if i == 1000:
                item = QTableWidgetItem("...")
            item.setFlags(item.flags() ^ Qt.ItemIsEditable)
            self.tableWidget_points.setItem(i, 0, item)
            if i == 1000:
                break
        if self.comboBox_loop_type.currentText() == "Value-List":
            for i in range(self.tableWidget_points.rowCount()):
                item = self.tableWidget_points.item(i, 0)
                item.setFlags(item.flags() | Qt.ItemIsEditable)

    def add_point(self):
        """Used to add a point when using the "Value-List" loop-type."""
        rows = self.tableWidget_points.rowCount()
        self.tableWidget_points.setRowCount(rows + 1)
        self.loop_step.val_list.append(0)
        item = QTableWidgetItem("0")
        self.tableWidget_points.setItem(rows, 0, item)
        item.setFlags(Qt.ItemIsEditable | Qt.ItemIsEnabled)

    def del_point(self):
        """Used to remove a point when using the "Value-List" loop-type."""
        indexes = self.tableWidget_points.selectedIndexes()
        for index in sorted(indexes, key=lambda x: -x.row()):
            self.tableWidget_points.removeRow(index.row())
            self.loop_step.val_list.pop(index.row())

    def value_list_changed(self, row):
        """Updates the `val_list` of the loopstep when it is changed.

        Parameters
        ----------
        row :


        Returns
        -------

        """
        if self.comboBox_loop_type.currentText() == "Value-List":
            try:
                self.loop_step.val_list[row] = float(
                    self.tableWidget_points.item(row, 0).text()
                )
            except ValueError:
                return



"""This module provides functions for communicating with NOMAD."""

import os.path
import requests
from PySide6.QtWidgets import QDialog

from nomad_camels.nomad_integration.nomad_login import LoginDialog
from nomad_camels.utility.dict_recursive_string import dict_recursive_string
from nomad_camels.utility import variables_handling
import re
import logging


def correct_timestamp(file_path):
    """Corrects the timestamp in the file path to be compatible with NOMAD. Replaces the plus symbol with p."""
    # Define the regex for the timestamp
    timestamp_regex = r"\d{4}-\d{2}-\d{2}T\d{2}-\d{2}-\d{2}_\d{6}\+\d{2}-\d{2}"

    # Find the timestamp in the file path
    match = re.search(timestamp_regex, file_path)

    if match:
        # Extract the timestamp
        timestamp = match.group()

        # Replace the plus symbol with "p"
        timestamp_with_p = timestamp.replace("+", "p")

        # Replace the timestamp in the file path with the new timestamp
        new_file_path = file_path.replace(timestamp, timestamp_with_p)

        return new_file_path

    else:
        return file_path


# by default the url of the central NOMAD at first
central_url = "http://nomad-lab.eu/prod/v1/staging/api/v1"
nomad_url = ""

token = ""
auth = {"Authorization": f"Bearer {token}"}


def login_to_nomad(parent=None):
    """Opens a login dialog and uses the resulting username/password to login to
    NOMAD. The received authentication token is saved to the module's global
    variable `token`.

    Parameters
    ----------
    parent : QWidget
        the parent widget for the login dialog

    Returns
    -------
    the authentication token
    """
    global token, auth, nomad_url
    dialog = LoginDialog(parent)
    if dialog.exec() != QDialog.Accepted:
        return
    nomad_url = dialog.url
    if not nomad_url:
        nomad_url = central_url
    make_correct_url()
    if dialog.token:
        local_auth = {"Authorization": f"Bearer {dialog.token}"}
        response = requests.get(f"{nomad_url}/auth/signature_token", headers=local_auth)
        check_response(response, "Login failed!")
        token = dialog.token
        auth = local_auth
    else:
        login = {"username": dialog.username, "password": dialog.password}
        response = requests.get(f"{nomad_url}/auth/token", params=login)
        check_response(response, "Login failed!")
        token = response.json()["access_token"]
        auth = {"Authorization": f"Bearer {token}"}
    return token


def ensure_login(parent=None):
    """Is always the first function to be called. Thus this function updates
    possible changes in the used NOMAD URL. If the URL is different to the last
    used one, `logout_of_nomad` is called. Then, if `token` is empty,
    `login_to_nomad` is called.

    Parameters
    ----------
    parent : QWidget
        the parent widget for the login dialog of `login_to_nomad`"""
    global nomad_url
    if "NOMAD_URL" in variables_handling.preferences:
        url = variables_handling.preferences["NOMAD_URL"]
        if url and url != nomad_url and nomad_url != central_url:
            nomad_url = url
            logout_of_nomad()
    make_correct_url()
    if not token:
        login_to_nomad(parent)


def make_correct_url(url=None):
    if url is None:
        global nomad_url
    else:
        nomad_url = url
    if "/gui/" in nomad_url:
        nomad_url = nomad_url.split("/gui/")[0]
    elif nomad_url.endswith("/gui"):
        nomad_url = nomad_url.split("/gui")[0]
    if nomad_url.endswith("/"):
        nomad_url = nomad_url[:-1]
    if not nomad_url.endswith("/api/v1"):
        nomad_url += "/api/v1"
    return nomad_url


def check_response(response, fail_info=""):
    """Checks the given `response` from NOMAD and raises an Exception if the
    request did not work.

    Parameters
    ----------
    response : requests.Response
        Response from the server. Its status code will be checked.
    fail_info : str
        (Default value = '')
        If this string is not empty, it will display as the first part of the
        raised exception. Useful to specify, what went wrong.
    """
    if response.status_code != 200:
        info = response.json()
        if fail_info:
            except_string = f"{fail_info}\n"
        else:
            except_string = ""
        except_string += dict_recursive_string(info)
        raise Exception(except_string)


def logout_of_nomad():
    """Resets the global `token` to an empty string, thus logging out of NOMAD."""
    global token, auth
    token = ""
    auth = {"Authorization": f"Bearer {token}"}


def get_user_uploads(parent=None):
    """Provides the uploads of the logged in user. If there is no login token, `login_to_nomad` will be called.

    Parameters
    ----------
    parent : QWidget
        the parent widget for the login dialog of `login_to_nomad`

    Returns
    -------
    the user's uploads
    """
    ensure_login(parent)
    response = requests.get(f"{nomad_url}/uploads", headers=auth)
    check_response(response, "Could not get uploads!")
    try:
        response_json = response.json()
    except requests.exceptions.JSONDecodeError as e:
        logging.error(f"Failed to decode JSON response: {e}")
        logging.error(f"Response content: {response.content}")
        raise e
    return response_json.get("data", [])


def get_user_upload_names(parent=None):
    """Provides a list of the uploads belonging to the logged in user by their
    name.

    Parameters
    ----------
    parent : QWidget
        the parent widget for the login dialog of `login_to_nomad`

    Returns
    -------
    list[str]
        the names of the uploads belonging to the user
    """
    uploads = get_user_uploads(parent)
    upload_names = []
    for upload in uploads:
        if "upload_name" in upload:
            upload_names.append(upload["upload_name"])
        else:
            upload_names.append(upload["upload_id"])
    return upload_names


def upload_file(
    file, upload_name, upload_path="CAMELS_data", overwrite_if_exists=True, parent=None
):
    """
    Uploads a file into a given NOMAD upload.

    Parameters
    ----------
    file : str, path
        the path of the file that should be uploaded
    upload_name : str
        the upload_name or upload_id of the upload, where the file should go
    upload_path : str
        (Default value = 'CAMELS_data')
        the path inside the upload on NOMAD where the file should go
    overwrite_if_exists : bool
        (Default value = True)
        if set to True , overwrites the file if it already exists
    parent : QWidget
        the parent widget for the login dialog of `login_to_nomad` (if needed)
    """
    uploads = get_user_uploads(parent)
    upload_id = None
    for upload in uploads:
        if "upload_name" not in upload and upload["upload_id"] == upload_name:
            upload_id = upload_name
            break
        if "upload_name" in upload and upload["upload_name"] == upload_name:
            upload_id = upload["upload_id"]
            break
    if not upload_id:
        raise Exception(f"Could not find upload {upload_name}!")
    file_corrected_timestamp = correct_timestamp(file)
    params = {
        "overwrite_if_exists": "true" if overwrite_if_exists else "false",
        "file_name": os.path.basename(file_corrected_timestamp),
    }
    head = {"accept": "application/json"}
    head.update(auth)
    with open(file, "rb") as f:
        response = requests.put(
            f"{nomad_url}/uploads/{upload_id}/raw/{upload_path}",
            data=f,
            headers=head,
            params=params,
        )
    check_response(response, "Failed to upload to NOMAD!")
    return response


def get_user_information(parent=None):
    """Retrieves the information of the logged in user.

    Parameters
    ----------
    parent : QWidget
        the parent widget for the login dialog of `login_to_nomad` (if needed)

    Returns
    -------
    the response's data, i.e. the user information
    """
    ensure_login(parent)
    response = requests.get(f"{nomad_url}/users/me", headers=auth)
    check_response(response, "Could not get user information from NOMAD")
    return response.json()


def get_entries(parent=None, owner="user"):
    """Retrieves the entries of the currently logged in user.

    Parameters
    ----------
    parent : QWidget
        the parent widget for the login dialog of `login_to_nomad` (if needed)
    owner : str
        The owner of the entries. Allowed values are 'public', 'all', 'visible',
        'shared', 'user', 'staging', 'admin'. Default is 'user'.
    """
    ensure_login(parent)
    params = {"owner": owner, "page_size": 1000}
    head = {"accept": "application/json"}
    head.update(auth)
    response = requests.get(f"{nomad_url}/entries/archive", headers=head, params=params)
    check_response(response, "Could not retrieve entry-information from NOMAD")
    return response.json()



from PySide6.QtWidgets import QLabel, QDialog, QGridLayout, QLineEdit, QDialogButtonBox

import hashlib

from nomad_camels.ui_widgets.warn_popup import WarnPopup


class Password_Dialog(QDialog):
    """
    Dialog to enter a password or create a new one.
    If double_pass is True or compare_hash is None, after accepting the dialog, the entered password can be accessed via the password attribute.

    Parameters
    ----------
    parent : QWidget
        Parent widget.
    double_pass : bool, optional
        Whether to ask for a password twice, by default False.
        Use this, if the user should create a new password.
    compare_hash : str, optional
        Hashed password to compare with, by default None.
        Use this, if the entered password should be compared directly.
    """

    def __init__(self, parent=None, double_pass=False, compare_hash=None):
        super().__init__(parent)
        self.compare_hash = compare_hash
        self.setWindowTitle("Enter Password - NOMAD CAMELS")
        self.double_pass = double_pass
        # if double_pass:
        self.label_text = QLabel("Enter password:")
        # else:
        #     self.label_text = QLabel('Enter password to save changes:')
        self.label_2 = QLabel("Confirm password:")
        self.lineEdit_password = QLineEdit()
        self.lineEdit_password.setEchoMode(QLineEdit.Password)
        self.lineEdit_password_2 = QLineEdit()
        self.lineEdit_password_2.setEchoMode(QLineEdit.Password)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)

        self.layout = QGridLayout()
        self.layout.addWidget(self.label_text, 0, 0)
        self.layout.addWidget(self.lineEdit_password, 0, 1)
        if double_pass:
            self.layout.addWidget(self.label_2, 1, 0)
            self.layout.addWidget(self.lineEdit_password_2, 1, 1)
        self.layout.addWidget(self.buttonbox, 2, 0, 1, 2)
        self.setLayout(self.layout)

        self.password = None

    def accept(self):
        if self.double_pass:
            if self.lineEdit_password.text() == self.lineEdit_password_2.text():
                self.password = self.lineEdit_password.text()
                return super().accept()
            else:
                self.lineEdit_password.setText("")
                self.lineEdit_password_2.setText("")
                WarnPopup(
                    text="Passwords did not match", title="Passwords did not match"
                )
                return
        password = self.lineEdit_password.text()
        if self.compare_hash:
            hashed_password = hash_password(password)
            if hashed_password != self.compare_hash:
                self.lineEdit_password.setText("")
                WarnPopup(
                    text="Incorrect password, please try again",
                    title="Incorrect password",
                )
                return
        else:
            self.password = self.lineEdit_password.text()
        return super().accept()


def hash_password(password):
    return hashlib.sha256(password.encode()).hexdigest()



import yaml
import jsonschema
from jsonschema import validate
from openapi_schema_validator import OAS30Validator

# Load and parse the OpenAPI YAML file
with open(r'nomad_camels\api\api_schema.yaml', 'r') as file:
    openapi_spec = yaml.safe_load(file)

# Validate the OpenAPI 3.0.3 specification
oas30_validator = OAS30Validator(openapi_spec)
oas30_validator.validate(openapi_spec)

# Extract the schema for Protocols
protocols_schema = openapi_spec['components']['schemas']['Protocols']

# Example of valid input data
valid_data = {
    "name": "Example Protocol",
    "JSON_definition": {"key": "value"}
}

# Example of invalid input data (missing JSON_definition)
invalid_data = {
    "name": "Example Protocol"
    # Missing "JSON_definition"
}

# Example of invalid input data (wrong type for JSON_definition)
invalid_data_wrong_type = {
    "name": "Example Protocol",
    "JSON_definition": "this should be a dict"
}

# Function to validate input data
def validate_protocol(data):
    try:
        validate(instance=data, schema=protocols_schema)
        print("Validation successful!")
    except jsonschema.exceptions.ValidationError as e:
        print(f"Validation error: {e.message}")

# Validate the example data
validate_protocol(valid_data)  # This should print "Validation successful!"
validate_protocol(invalid_data)  # This should print a validation error message
validate_protocol(invalid_data_wrong_type)  # This should print a validation error message



from PySide6.QtWidgets import QWidget, QLabel, QGridLayout, QComboBox, QCheckBox

from nomad_camels.main_classes.loop_step import Loop_Step_Config, Loop_Step
from nomad_camels.ui_widgets.variable_tool_tip_box import Variable_Box
from nomad_camels.ui_widgets.channels_check_table import Channels_Check_Table

wait_types = ["simple wait", "wait with progress bar", "wait for condition"]


class Wait_Loop_Step(Loop_Step):
    """
    A loopstep to simply wait some defined time.

    Attributes
    ----------
    wait_time : str, float
        The how long the protocol execution should pause in seconds.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Wait"
        if step_info is None:
            step_info = {}
        self.wait_time = step_info["wait_time"] if "wait_time" in step_info else 0.0
        self.wait_type = (
            step_info["wait_type"] if "wait_type" in step_info else "simple wait"
        )
        self.skipable = step_info["skipable"] if "skipable" in step_info else True
        self.condition = step_info["condition"] if "condition" in step_info else ""
        self.read_channels = (
            step_info["read_channels"] if "read_channels" in step_info else []
        )

    def get_add_main_string(self):
        add_main_string = super().get_add_main_string()
        if self.wait_type == "wait with progress bar":
            add_main_string += f'\tboxes["bar_{self.name}"] = helper_functions.Waiting_Bar(title="{self.name} waiting...", skipable={self.skipable}, with_timer=True)\n'
        elif self.wait_type == "wait for condition":
            from nomad_camels.bluesky_handling import builder_helper_functions

            add_main_string += builder_helper_functions.get_plot_add_string(
                self.name, f'"{self.full_name}"'
            )
        return add_main_string

    def get_outer_string(self):
        outer_string = super().get_outer_string()
        if self.wait_type == "wait for condition":
            from nomad_camels.bluesky_handling import builder_helper_functions
            from nomad_camels.frontpanels.plot_definer import Plot_Info

            y_axes = [self.condition]
            for comparison in ["<=", ">=", "<", ">", "=="]:
                if comparison in self.condition:
                    y_axes += self.condition.split(comparison)
                    break
            plot_data = [
                Plot_Info(
                    "Value-List", title=self.full_name, y_axes={"formula": y_axes}
                )
            ]
            outer_string += builder_helper_functions.plot_creator(
                plot_data,
                f"create_plots_{self.name}",
                plot_is_box=True,
                box_names=f"bar_{self.name}",
                skip_box=self.skipable,
            )[0]
        return outer_string

    def update_used_devices(self):
        """All devices that should be read are added to the used_devices."""
        self.used_devices = []
        if not self.wait_type == "wait for condition":
            return
        from nomad_camels.utility import variables_handling

        for channel in variables_handling.channels:
            if channel in self.read_channels:
                device = variables_handling.channels[channel].device
                if device not in self.used_devices:
                    self.used_devices.append(device)

    def get_protocol_string(self, n_tabs=1):
        """The protocol just calls `bps.wait(`wait_time`)`, where `wait_time` is
        evaluated by the protocol's evaluator."""
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        if self.wait_type == "simple wait":
            protocol_string += (
                f'{tabs}yield from bps.sleep(eva.eval("{self.wait_time}"))\n'
            )
        elif self.wait_type == "wait with progress bar":
            protocol_string += f'{tabs}boxes["bar_{self.name}"].skip = False\n'
            protocol_string += (
                f'{tabs}boxes["bar_{self.name}"].helper.executor.emit()\n'
            )
            protocol_string += (
                f"{tabs}bar_{self.name}_delta_t = eva.eval('{self.wait_time} / 100')\n"
            )
            protocol_string += f'{tabs}boxes["bar_{self.name}"].setter.set_wait_time(eva.eval("{self.wait_time}"))\n'
            protocol_string += f'{tabs}while not boxes["bar_{self.name}"].skip and not boxes["bar_{self.name}"].setter.timer > eva.eval("{self.wait_time}"):\n'
            protocol_string += (
                f"{tabs}\tyield from bps.sleep(bar_{self.name}_delta_t)\n"
            )
            protocol_string += (
                f'{tabs}\tboxes["bar_{self.name}"].setter.update_timer()\n'
            )
            protocol_string += (
                f'{tabs}boxes["bar_{self.name}"].setter.hide_signal.emit()\n'
            )
        elif self.wait_type == "wait for condition":
            protocol_string += self.get_channels_string(tabs)
            protocol_string += f"{tabs}yield from bps.trigger_and_read(channels_{self.variable_name()}, name='{self.full_name}')\n"
            protocol_string += (
                f'{tabs}boxes["bar_{self.name}_0"].helper.executor.emit()\n'
            )
            protocol_string += f'{tabs}while not eva.eval("{self.condition}") and not boxes["bar_{self.name}_0"].skip:\n'
            protocol_string += (
                f'{tabs}\tyield from bps.sleep(eva.eval("{self.wait_time}"))\n'
            )
            protocol_string += f"{tabs}\tyield from bps.trigger_and_read(channels_{self.variable_name()}, name='{self.full_name}')\n"
            protocol_string += (
                f'{tabs}boxes["bar_{self.name}_0"].setter.hide_signal.emit()\n'
            )
        return protocol_string

    def variable_name(self):
        """Returns the name of this step as a valid variable name, to specify
        the channels for this read."""
        from nomad_camels.utility import fit_variable_renaming

        return fit_variable_renaming.replace_name(self.name)

    def get_channels_string(self, tabs):
        """
        Gives a string of the channels that should be read. This may also be
        used by the Trigger_Channels step.

        Parameters
        ----------
        tabs : str
            A string including the tabs for intendation.
        """
        from nomad_camels.loop_steps.read_channels import get_channel_string

        channel_string = f"{tabs}channels_{self.variable_name()} = ["
        if not self.read_channels:
            raise Exception(f"Trying to read no channel in {self.full_name}!")
        for channel in self.read_channels:
            channel_string += get_channel_string(channel)
        channel_string = channel_string[:-2] + "]\n"
        return channel_string

    def get_protocol_short_string(self, n_tabs=0):
        """Tells the wait time."""
        short_string = super().get_protocol_short_string(n_tabs)
        if self.wait_type != "wait for condition":
            short_string = f"{short_string[:-1]} - {self.wait_time} s\n"
        else:
            short_string = f"{short_string[:-1]} - '{self.condition}'\n"
        return short_string


class Wait_Loop_Step_Config(Loop_Step_Config):
    """The configuration just provides a line to enter the time to wait."""

    def __init__(self, loop_step: Wait_Loop_Step, parent=None):
        super().__init__(parent, loop_step)
        self.sub_widget = Wait_Loop_Step_Config_Sub(loop_step, self)
        self.layout().addWidget(self.sub_widget, 1, 0, 1, 5)

    def update_step_config(self):
        """ """
        super().update_step_config()
        self.loop_step.wait_time = self.sub_widget.lineEdit_duration.text()
        self.loop_step.wait_type = self.sub_widget.comboBox_type.currentText()
        self.loop_step.skipable = self.sub_widget.checkbox_skipable.isChecked()
        self.loop_step.condition = self.sub_widget.condition_line.text()
        self.loop_step.read_channels = self.sub_widget.read_channels_table.get_info()[
            "channel"
        ]


class Wait_Loop_Step_Config_Sub(QWidget):
    """The QLineEdit and labels to make everything clear are provided."""

    def __init__(self, loop_step: Wait_Loop_Step, parent=None):
        super().__init__(parent)
        self.loop_step = loop_step

        label_type = QLabel("Wait type")
        self.comboBox_type = QComboBox(self)
        self.comboBox_type.addItems(wait_types)
        self.comboBox_type.setCurrentText(loop_step.wait_type)
        self.comboBox_type.currentTextChanged.connect(self.type_changed)

        self.label1 = QLabel("Wait for")
        self.label2 = QLabel("seconds")
        self.lineEdit_duration = Variable_Box(self)
        self.lineEdit_duration.setText(str(loop_step.wait_time))
        self.lineEdit_duration.textChanged.connect(self.update_duration)

        self.checkbox_skipable = QCheckBox("skipable")
        self.checkbox_skipable.setChecked(loop_step.skipable)
        self.checkbox_skipable.setToolTip("The progress bar has a skip button.")

        self.label_condition = QLabel("Wait until:")
        self.condition_line = Variable_Box(self)
        self.condition_line.setText(loop_step.condition)
        self.condition_line.setToolTip("Wait until this condition is fulfilled.")

        info_dict = {"channel": loop_step.read_channels}
        self.read_channels_table = Channels_Check_Table(
            self,
            ["read?", "channel"],
            info_dict=info_dict,
            title="Read channels",
        )

        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(label_type, 0, 0)
        layout.addWidget(self.comboBox_type, 0, 1, 1, 2)
        layout.addWidget(self.label1, 1, 0)
        layout.addWidget(self.lineEdit_duration, 1, 1)
        layout.addWidget(self.label2, 1, 2)
        layout.addWidget(self.checkbox_skipable, 2, 0, 1, 3)
        layout.addWidget(self.label_condition, 3, 0)
        layout.addWidget(self.condition_line, 3, 1, 1, 2)
        layout.addWidget(self.read_channels_table, 4, 0, 1, 3)
        self.setLayout(layout)
        self.type_changed()

    def update_duration(self):
        """ """
        self.loop_step.wait_time = self.lineEdit_duration.text()

    def type_changed(self):
        """ """
        wait_type = self.comboBox_type.currentText()
        if wait_type == "simple wait":
            self.checkbox_skipable.setHidden(True)
        else:
            self.checkbox_skipable.setHidden(False)
        if wait_type == "wait for condition":
            self.read_channels_table.setHidden(False)
            self.condition_line.setHidden(False)
            self.label_condition.setHidden(False)
            self.label1.setText("read every")
        else:
            self.read_channels_table.setHidden(True)
            self.condition_line.setHidden(True)
            self.label_condition.setHidden(True)
            self.label1.setText("Wait for")



# %%
import requests
from requests.auth import HTTPBasicAuth
import json

api_key = "5rCMKfU4JJV6SjOy_z237lk0-OoxdQIM_MjOPu2P2BGLENG60jY1eA"
# %%
# result_no_key = requests.get("http://127.0.0.1:5000/get_protocols")
# print(result_no_key.text)
# %%

result = requests.get("http://localhost:5000/protocols", auth=("", f"{api_key}"))
print(result.text)



"""This module provides a way to produce all possible protocol steps from the UI"""

from nomad_camels.main_classes.loop_step import Loop_Step
from nomad_camels.loop_steps import (
    for_while_loops,
    read_channels,
    set_channels,
    wait_loop_step,
    if_step,
    prompt_loop_step,
    run_subprotocol,
    simple_sweep,
    gradient_descent,
    change_device_config,
    nd_sweep,
    set_variables,
    set_value_popup,
    export_data,
    call_function,
    execute_python_file,
    api_call,
)

from nomad_camels.utility import variables_handling

step_type_config = {
    "Call Function": [call_function.Call_Function, call_function.Call_Function_Config],
    "Change Device Config": [
        change_device_config.Change_DeviceConf,
        change_device_config.Change_DeviceConf_Config,
    ],
    # 'Export Data': [export_data.Export_Data_Step,
    #                 export_data.Export_Data_Step_Config],
    "For Loop": [for_while_loops.For_Loop_Step, for_while_loops.For_Loop_Step_Config],
    "Gradient Descent": [
        gradient_descent.Gradient_Descent_Step,
        gradient_descent.Gradient_Descent_Config,
    ],
    "If": [if_step.If_Loop_Step, if_step.If_Step_Config],
    "ND Sweep": [nd_sweep.ND_Sweep, nd_sweep.ND_Sweep_Config],
    "Prompt": [
        prompt_loop_step.Prompt_Loop_Step,
        prompt_loop_step.Prompt_Loop_Step_Config,
    ],
    "Read Channels": [read_channels.Read_Channels, read_channels.Read_Channels_Config],
    "Run Subprotocol": [
        run_subprotocol.Run_Subprotocol,
        run_subprotocol.Run_Subprotocol_Config,
    ],
    "Set Channels": [set_channels.Set_Channels, set_channels.Set_Channels_Config],
    "Set Value Popup": [
        set_value_popup.Set_Value_Popup,
        set_value_popup.Set_Value_Popup_Config,
    ],
    "Set Variables": [set_variables.Set_Variables, set_variables.Set_Variables_Config],
    "Simple Sweep": [simple_sweep.Simple_Sweep, simple_sweep.Simple_Sweep_Config],
    "Trigger Channels": [
        read_channels.Trigger_Channels_Step,
        read_channels.Trigger_Channels_Config,
    ],
    "While Loop": [
        for_while_loops.While_Loop_Step,
        for_while_loops.While_Loop_Step_Config,
    ],
    "Wait": [wait_loop_step.Wait_Loop_Step, wait_loop_step.Wait_Loop_Step_Config],
    "Execute Python File": [
        execute_python_file.Execute_Python_File,
        execute_python_file.Execute_Python_File_Config,
    ],
    "API Call": [
        api_call.API_Call,
        api_call.API_Call_Config,
    ],
}

steps_with_children = []
for step in step_type_config:
    if step_type_config[step][0]().has_children:
        steps_with_children.append(step)

non_addables = {
    "If_Sub": [if_step.If_Sub_Step, if_step.Sub_Step_Config],
    "Elif_Sub": [if_step.Elif_Sub_Step, if_step.Sub_Step_Config],
    "Else_Sub": [if_step.Else_Sub_Step, if_step.Sub_Step_Config],
}


def get_device_steps():
    """Goes through all the devices and checks, whether they provide
    their own types of loop-steps.

    Returns
    -------
    device_steps
        dictionary of the possible steps provided by the devices
    """
    device_steps = {}
    for devname, device in sorted(
        variables_handling.devices.items(), key=lambda x: x[0].lower()
    ):
        device_steps.update(device.get_special_steps())
    return device_steps


def make_step(step_type, step_info=None, children=None, protocol=None):
    """Checks whether the given `step_type` is supported, if yes, then a
    step of that type with the given info and children will be created.
    If there is no step_info, a new step is made.

    Parameters
    ----------
    step_type : str
        A string representing the type of the step. Should correspond to one of
        the keys of `step_type_config` or a device specific step.
    step_info : dict, None
        (Default value = None)
        The information for the step's configuration
    children : list[Loop_Step], None
        (Default value = None)
        The produced step's children. Handed over to it's constructor.
    protocol : Measurement_Protocol
        (Default value = None)
        The protocol, in which the step should be.

    Returns
    -------
    step
        The produced step.
    """
    dev_steps = get_device_steps()
    if step_type in step_type_config:
        if step_info is None:
            name = step_type.replace(" ", "_")
        else:
            name = step_info["name"]
        return step_type_config[step_type][0](
            name=name, step_info=step_info, children=children, protocol=protocol
        )
    elif step_type in dev_steps:
        if step_info is None:
            name = step_type.replace(" ", "_")
        else:
            name = step_info["name"]
        return dev_steps[step_type][0](
            name=name, step_info=step_info, children=children, protocol=protocol
        )
    elif step_type in non_addables:
        if step_info is None:
            name = step_type.replace(" ", "_")
        else:
            name = step_info["name"]
        return non_addables[step_type][0](
            name=name, step_info=step_info, children=children, protocol=protocol
        )
    return Loop_Step(name="fail")


def get_config(step: Loop_Step):
    """Returns the Loop_Step_Config belonging to the given step.

    Parameters
    ----------
    step : Loop_Step
        The step for which the config-widget should be opened.

    Returns
    -------
    step_config
        The Loop_Step_Config corresponding to the given `step`.
    """
    step_type = step.step_type
    dev_steps = get_device_steps()
    if step_type in step_type_config:
        return step_type_config[step_type][1](loop_step=step)
    elif step_type in dev_steps:
        return dev_steps[step_type][1](loop_step=step)
    elif step_type in non_addables:
        return non_addables[step_type][1](loop_step=step)
    raise Exception("Loop Step configuration is not defined!")



import numpy as np

preferences = {}


def format_number(value):
    """Format a number-string the way it is specified in the preferences.

    Parameters
    ----------
    value : float, int
        The number that should be displayed as string.

    Returns
    -------
    str
        The formatted string of the number.
    """
    if isinstance(value, int):
        return str(value)
    if isinstance(value, str):
        try:
            value = float(value)
        except ValueError:
            return value
    if "number_format" not in preferences or "n_decimals" not in preferences:
        return str(f"{value:.2f}")
    if preferences["number_format"] == "plain":
        return f'{value:.{preferences["n_decimals"]}f}'
    elif preferences["number_format"] == "scientific":
        return f'{value:.{preferences["n_decimals"]}e}'
    else:
        if (
            np.abs(value) >= 10 ** preferences["mixed_from"]
            or np.abs(value) < 10 ** (-preferences["mixed_from"] + 1)
        ) and value != 0:
            return f'{value:.{preferences["n_decimals"]}e}'
        return f'{value:.{preferences["n_decimals"]}f}'



"""This module takes care of all the settings regarding logging to a logfile.
The logfile is put into the `appdata_path`, a filehandler with specific
formatting is added to logging's app_log"""

import logging
import os
from logging.handlers import RotatingFileHandler

from nomad_camels.utility import load_save_functions
from nomad_camels.utility import variables_handling
from nomad_camels.utility.load_save_functions import standard_pref

log_levels = {
    "All": logging.NOTSET,
    "Debug": logging.DEBUG,
    "Info": logging.INFO,
    "Warning": logging.WARNING,
    "Error": logging.ERROR,
    "Critical": logging.CRITICAL,
}

my_handler = None


def update_log_settings():
    """Reads the logging settings from the current preferences and sets them for
    the logfile-handler. Including: the level of logging, the maximum size for
    the logfile, the number of old logfiles, when one is full."""
    global my_handler
    logfile = os.path.join(load_save_functions.appdata_path, "nomad_camels.log")
    if not os.path.isfile(logfile):
        if not os.path.isdir(load_save_functions.appdata_path):
            os.makedirs(load_save_functions.appdata_path)
        with open(logfile, "w") as f:
            pass
    app_log = logging.getLogger("root")
    if my_handler is None:
        my_handler = RotatingFileHandler(
            logfile,
            mode="a",
            maxBytes=standard_pref["logfile_size"] * 1024 * 1024,
            backupCount=standard_pref["logfile_backups"],
        )

        log_formatter = logging.Formatter(
            "%(asctime)s %(levelname)s %(funcName)s(%(lineno)d) %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        my_handler.setFormatter(log_formatter)

        app_log.addHandler(my_handler)

    my_handler.baseFilename = logfile

    prefs = variables_handling.preferences
    if "log_level" in prefs and prefs["log_level"]:
        log_level = prefs["log_level"]
    else:
        log_level = standard_pref["log_level"]
    if "logfile_size" in prefs and prefs["logfile_size"]:
        logfile_size = prefs["logfile_size"]
    else:
        logfile_size = standard_pref["logfile_size"]
    if "logfile_backups" in prefs:
        logfile_backups = prefs["logfile_backups"]
    else:
        logfile_backups = standard_pref["logfile_backups"]

    my_handler.backupCount = logfile_backups
    my_handler.maxBytes = logfile_size * 1024 * 1024
    if log_level in log_levels:
        log_lv = log_levels[log_level]
    else:
        log_lv = logging.NOTSET
    my_handler.setLevel(log_lv)
    app_log.setLevel(log_lv)


update_log_settings()



def dict_recursive_string(data, indent=0):
    """Recursively adds the given dictionary `data` to a string in the format
    "key: value" where each key-value pair has its own line and nested
    dictionaries are indented. Returns the resulting string."""
    rec_string = ""
    for key, value in data.items():
        if isinstance(value, dict):
            rec_string += f"{' ' * 4 * indent}{key}:"
            rec_string += dict_recursive_string(value, indent + 1)
        else:
            rec_string += f"{' ' * 4 * indent}{key}: {value}\n"
    return rec_string



from nomad_camels.gui.enterTextDialog import Ui_EnterTextDialog
from PySide6.QtWidgets import QDialog


class EnterTextDialog(Ui_EnterTextDialog, QDialog):
    """A simple QDialog-Box that has the value of the entered textbox in "self.value".

    - parent: parent-widget, passed to QDialog
    - window_title: The title of the pop-up window
    - label: The label shown in front of the QLineEdit

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, parent=None, window_title="", label=""):
        super(EnterTextDialog, self).__init__(parent)
        self.setupUi(self)
        self.setWindowTitle(window_title)
        self.label.setText(label)
        self.value = None

    def accept(self) -> None:
        """ """
        self.value = self.lineEdit_text.text()
        super().accept()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'fit_definer.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QRadioButton,
    QSizePolicy,
    QSpacerItem,
    QWidget,
)


class Ui_Fit_Definer(object):
    def setupUi(self, Fit_Definer):
        if not Fit_Definer.objectName():
            Fit_Definer.setObjectName("Fit_Definer")
        Fit_Definer.resize(287, 243)
        self.gridLayout = QGridLayout(Fit_Definer)
        self.gridLayout.setObjectName("gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.radioButton_custom_func = QRadioButton(Fit_Definer)
        self.radioButton_custom_func.setObjectName("radioButton_custom_func")

        self.gridLayout.addWidget(self.radioButton_custom_func, 1, 1, 1, 1)

        self.radioButton_predef_func = QRadioButton(Fit_Definer)
        self.radioButton_predef_func.setObjectName("radioButton_predef_func")
        self.radioButton_predef_func.setChecked(True)

        self.gridLayout.addWidget(self.radioButton_predef_func, 1, 0, 1, 1)

        self.lineEdit_custom_func = QLineEdit(Fit_Definer)
        self.lineEdit_custom_func.setObjectName("lineEdit_custom_func")

        self.gridLayout.addWidget(self.lineEdit_custom_func, 2, 1, 1, 1)

        self.comboBox_predef_func = QComboBox(Fit_Definer)
        self.comboBox_predef_func.setObjectName("comboBox_predef_func")

        self.gridLayout.addWidget(self.comboBox_predef_func, 2, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.gridLayout.addItem(self.verticalSpacer, 3, 0, 1, 1)

        self.label = QLabel(Fit_Definer)
        self.label.setObjectName("label")
        font = QFont()
        font.setPointSize(9)
        font.setBold(True)
        self.label.setFont(font)

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.checkBox_fit = QCheckBox(Fit_Definer)
        self.checkBox_fit.setObjectName("checkBox_fit")

        self.gridLayout.addWidget(self.checkBox_fit, 0, 1, 1, 1)

        self.checkBox_guess = QCheckBox(Fit_Definer)
        self.checkBox_guess.setObjectName("checkBox_guess")

        self.gridLayout.addWidget(self.checkBox_guess, 4, 0, 1, 1)

        self.checkBox_display_values = QCheckBox(Fit_Definer)
        self.checkBox_display_values.setObjectName("checkBox_display_values")

        self.gridLayout.addWidget(self.checkBox_display_values, 4, 1, 1, 1)

        self.retranslateUi(Fit_Definer)

        QMetaObject.connectSlotsByName(Fit_Definer)

    # setupUi

    def retranslateUi(self, Fit_Definer):
        Fit_Definer.setWindowTitle(
            QCoreApplication.translate("Fit_Definer", "Form", None)
        )
        self.radioButton_custom_func.setText(
            QCoreApplication.translate("Fit_Definer", "Custom function", None)
        )
        self.radioButton_predef_func.setText(
            QCoreApplication.translate("Fit_Definer", "Predefined function", None)
        )
        self.label.setText(
            QCoreApplication.translate("Fit_Definer", "Fit to: all y-axes", None)
        )
        self.checkBox_fit.setText(
            QCoreApplication.translate("Fit_Definer", "fit?", None)
        )
        self.checkBox_guess.setText(
            QCoreApplication.translate("Fit_Definer", "guess initial parameters", None)
        )
        self.checkBox_display_values.setText(
            QCoreApplication.translate("Fit_Definer", "display fit values", None)
        )

    # retranslateUi



from PySide6.QtWidgets import (
    QApplication,
    QLineEdit,
    QMenu,
    QDialog,
    QLabel,
    QWidget,
    QGridLayout,
    QPushButton,
    QScrollArea,
    QMainWindow,
    QTabWidget,
    QTabBar,
    QMessageBox,
    QComboBox,
)
from PySide6.QtGui import QDrag, QPixmap
from PySide6.QtCore import Qt, QMimeData, Signal, QByteArray, QMimeData

from collections import OrderedDict


class DragButton(QPushButton):
    """ """

    def __init__(self, text):
        super().__init__(text)

    def mouseMoveEvent(self, event):
        """

        Parameters
        ----------
        event :


        Returns
        -------

        """
        if event.buttons() == Qt.LeftButton:
            mimeData = QMimeData()
            drag = QDrag(self)
            drag.setMimeData(mimeData)
            drag.exec_(Qt.MoveAction)


class BidirectionalDict:
    """ """

    def __init__(self):
        self._forward = {}
        self._reverse = {}

    def __getitem__(self, key):
        return self._forward[key]

    def __setitem__(self, key, value):
        self._forward[key] = value
        self._reverse[value] = key

    def __delitem__(self, key):
        value = self._forward.pop(key)
        del self._reverse[value]

    def __len__(self):
        return len(self._forward)

    def __iter__(self):
        return iter(self._forward)

    def items(self):
        """ """
        return self._forward.items()

    def keys(self):
        """ """
        return self._forward.keys()

    def values(self):
        """ """
        return self._forward.values()

    def get_key(self, value):
        """

        Parameters
        ----------
        value :


        Returns
        -------

        """
        return self._reverse[value]

    def get_value(self, key):
        """

        Parameters
        ----------
        key :


        Returns
        -------

        """
        return self._forward[key]

    def pop(self, key):
        """

        Parameters
        ----------
        key :


        Returns
        -------

        """
        val = self._forward.pop(key)
        self._reverse.pop(val)
        return val

    def __repr__(self):
        return f"BidirectionalDict({self._forward})"

    def clear(self):
        self._forward.clear()
        self._reverse.clear()


class DropArea(QWidget):
    """ """

    order_changed = Signal(list)

    def __init__(self, button_wdith: int):
        super().__init__()
        # self.initUI()
        self.buttons = BidirectionalDict()
        self.button_order = []
        layout = QGridLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.setAcceptDrops(True)
        self.columnCount = 0
        self.rowCount = 0

        self.min_column_width = button_wdith + 1

    def dragEnterEvent(self, event):
        """

        Parameters
        ----------
        event :


        Returns
        -------

        """
        event.accept()

    def dropEvent(self, event):
        """

        Parameters
        ----------
        event :


        Returns
        -------

        """
        event.setDropAction(Qt.MoveAction)
        event.accept()

        source_widget = event.source()

        if not source_widget in self.buttons.values():
            return
        drag_pos = self.buttons.get_key(source_widget)
        drop_row, drop_col = self.get_drop_position(event.pos())
        if drop_row * self.columnCount + drop_col >= len(self.buttons):
            drop_col = int(len(self.buttons) - drop_row * self.columnCount - 1)
        child_at = self.layout().itemAtPosition(drop_row, drop_col).widget()
        drop_pos = self.buttons.get_key(child_at)

        if drag_pos == drop_pos:
            return
        adding = 0
        if self.button_order.index(drag_pos) < self.button_order.index(drop_pos):
            adding = 1
        self.button_order.remove(drag_pos)
        insert_pos = self.button_order.index(drop_pos)
        self.button_order.insert(insert_pos + adding, drag_pos)
        self.updateLayout()
        self.order_changed.emit(self.button_order)

    def get_drop_position(self, pos):
        """

        Parameters
        ----------
        pos :


        Returns
        -------

        """
        rect = self.rect()
        center = rect.center()
        width = self.width()
        height = self.height()
        x_dist = (pos - center).x() + width / 2
        y_dist = (pos - center).y() + height / 2
        column_width = width / self.columnCount
        row_height = height / self.rowCount
        return int(y_dist // row_height), int(x_dist // column_width)

    def updateLayout(self):
        """ """
        width = self.width()

        # calculate number of columns based on current width
        columns = max(1, width // self.min_column_width)

        # calculate new positions of buttons based on columns
        positions = [(i // columns, i % columns) for i in range(len(self.buttons))]
        self.columnCount = columns
        self.rowCount = len(positions) // columns + (
            1 if len(positions) % columns else 0
        )
        for button, position in zip(self.button_order, positions):
            self.layout().addWidget(self.buttons[button], *position)


class Drop_Scroll_Area(QScrollArea):
    """ """

    order_changed = Signal(list)

    def __init__(self, parent=None, button_width=120, button_height=120):
        super().__init__(parent=parent)
        self.drop_area = DropArea(button_width)
        self.setWidget(self.drop_area)
        self.setWidgetResizable(True)
        self.drop_area.updateLayout()
        self.button_width = button_width
        self.button_height = button_height
        self.setMinimumWidth(button_width)
        self.setMinimumHeight(button_height)
        self.drop_area.order_changed.connect(self.order_changed.emit)

    def resizeEvent(self, a0):
        """

        Parameters
        ----------
        a0 :


        Returns
        -------

        """
        super().resizeEvent(a0)
        self.updateLayout()

    def updateLayout(self):
        """ """
        self.drop_area.updateLayout()

    def add_button(self, button, name):
        """

        Parameters
        ----------
        button :

        name :


        Returns
        -------

        """
        button.setFixedSize(self.button_width, self.button_height)
        self.drop_area.buttons[name] = button
        self.drop_area.button_order.append(name)
        self.drop_area.updateLayout()

    def remove_button(self, name):
        """

        Parameters
        ----------
        name :


        Returns
        -------

        """
        button = self.drop_area.buttons.pop(name)
        self.drop_area.button_order.remove(name)
        self.drop_area.updateLayout()
        button.deleteLater()

    def clear_area(self):
        for button in self.drop_area.buttons:
            self.drop_area.buttons[button].deleteLater()
        self.drop_area.buttons.clear()
        self.drop_area.button_order.clear()

    def get_button_order(self):
        """ """
        return self.drop_area.button_order

    def rename_button(self, old_name, new_name):
        """

        Parameters
        ----------
        old_name :

        new_name :


        Returns
        -------

        """
        button = self.drop_area.buttons.pop(old_name)
        self.drop_area.buttons[new_name] = button
        ind = self.drop_area.button_order.index(old_name)
        self.drop_area.button_order.pop(ind)
        self.drop_area.button_order.insert(ind, new_name)
        button.rename(new_name)
        return button

    def disable_run_buttons(self):
        """ """
        for button in self.drop_area.buttons.values():
            button.small_button.setEnabled(False)

    def enable_run_buttons(self):
        """ """
        for button in self.drop_area.buttons.values():
            button.small_button.setEnabled(True)

    def disable_single_run(self, name):
        """

        Parameters
        ----------
        name :


        Returns
        -------

        """
        if name in self.drop_area.buttons:
            self.drop_area.buttons[name].small_button.setEnabled(False)

    def enable_single_run(self, name):
        """

        Parameters
        ----------
        name :


        Returns
        -------

        """
        if name in self.drop_area.buttons:
            self.drop_area.buttons[name].small_button.setEnabled(True)


class RenameTabWidget(QTabWidget):
    order_changed = Signal(list)

    def __init__(self, parent=None, tab_button_dict=None):
        super().__init__(parent)
        self.mainWindow = parent
        self.setAcceptDrops(True)
        self.tabBar().setMovable(True)

        # Add the "plus" tab
        self.plus_tab = QPushButton("+")
        self.plus_tab.setFlat(True)
        self.plus_tab.setFocusPolicy(Qt.NoFocus)
        self.plus_tab.setFixedSize(25, 25)
        self.plus_tab.clicked.connect(lambda x: self.create_new_tab())

        # Add the plus tab
        self.tab_button_dict = tab_button_dict or {}
        self.all_buttons = []
        self.editing_old_name = ""
        self.addPlusTab()

        # Add a context menu to the tab bar
        self.tabBar().setContextMenuPolicy(Qt.CustomContextMenu)
        self.tabBar().customContextMenuRequested.connect(self.context_menu)

    def showEvent(self, event):
        super().showEvent(event)
        self.setCurrentIndex(0)

    def addPlusTab(self):
        self.addTab(QWidget(), "")
        self.tabBar().setTabButton(self.count() - 1, QTabBar.RightSide, self.plus_tab)
        self.tabBar().setTabEnabled(self.count() - 1, False)
        self.plus_tab.setStyleSheet(
            """
            QPushButton {
                border: none; /* Remove border for a cleaner look */
                color: white; /* Text color */
                padding: 0px; /* Adjust padding as needed */
                padding-bottom: 5px;
                border-radius: 6px; /* Rounded corners with a radius of 10px */
                font-size: 22px;
                font-weight: bold;
                text-align: center;
                margin: 0px; /* Remove margin to avoid green box */
            }
            QPushButton:hover {
                color: #abaaa7; /* Slightly darker gray on hover */
            }
            QPushButton:pressed {
                color: #5e5e5c; /* Slightly darker gray when pressed */
            }
            """
        )

    def context_menu(self, pos):
        tab_index = self.tabBar().tabAt(pos)
        if tab_index == -1 or self.tabText(tab_index) == "":
            return
        menu = QMenu()
        menu.addAction("Add Tab", self.create_new_tab)
        menu.addAction("Remove Tab", lambda: self.tab_removing(tab_index))
        menu.addSeparator()
        menu.addAction("Rename Tab", lambda: self.rename_tab(tab_index))
        menu.exec(self.tabBar().mapToGlobal(pos))

    def tab_removing(self, index):
        if self.tabText(index) == "":
            return
        delete_question = QMessageBox.question(
            self,
            "Remove Tab",
            f"Are you sure you want to delete the tab '{self.tabText(index)}'?\nAll protocols inside the tab will be deleted as well!",
            QMessageBox.Yes | QMessageBox.No,
        )
        if delete_question != QMessageBox.Yes:
            return
        for protocol_ in self.tab_button_dict[self.tabText(index)]:
            del self.tab_button_dict[protocol_]
        del self.tab_button_dict[self.tabText(index)]
        self.removeTab(index)

    def rename_tab(self, index):
        if self.tabText(index) == "":
            return
        old_name = self.tabText(index)
        line_edit = QLineEdit(old_name)
        line_edit.selectAll()
        line_edit.editingFinished.connect(
            lambda: self.handle_editing_finished(index, line_edit)
        )
        self.tabBar().setTabText(index, "")
        self.tabBar().setTabButton(index, QTabBar.RightSide, line_edit)
        self.editing_old_name = old_name
        line_edit.setFocus()

    def handle_editing_finished(self, index, line_edit):
        old_name = self.editing_old_name
        new_name = line_edit.text()
        if new_name:
            self.tabBar().setTabText(index, new_name)
        else:
            self.tabBar().setTabText(index, old_name)
            new_name = old_name
        self.tabBar().setTabButton(index, QTabBar.RightSide, None)
        self.tab_button_dict[new_name] = self.tab_button_dict.pop(old_name)

    def updateLayout(self):
        for tab in range(self.count() - 1):  # Skip the last tab (plus tab)
            self.widget(tab).updateLayout()

    def get_tab_by_name(self, name, make_new=False):
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            if self.tabText(i) == name:
                return self.widget(i)
        if make_new:
            tab = Drop_Scroll_Area()
            self.insertTab(self.count() - 1, tab, name)  # Insert before the plus tab
            return tab
        return None

    def add_button(self, button, name, tab_name=""):
        self.all_buttons.append(button)
        widget = self.get_tab_by_name(tab_name, make_new=True)
        widget.add_button(button, name)

    def remove_button(self, name):
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            try:
                self.widget(i).remove_button(name)
            except KeyError:
                pass

    def clear_area(self):
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            self.widget(i).clear_area()
        # remove all tabs except the plus tab
        while self.count() > 1:
            self.removeTab(0)

    def get_button_order(self, tab_name=""):
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            if self.tabText(i) == tab_name:
                return self.widget(i).get_button_order()
        return []

    def rename_button(self, old_name, new_name):
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            try:
                return self.widget(i).rename_button(old_name, new_name)
            except KeyError:
                pass

    def disable_run_buttons(self):
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            self.widget(i).disable_run_buttons()

    def enable_run_buttons(self):
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            self.widget(i).enable_run_buttons()

    def disable_single_run(self, name):
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            self.widget(i).disable_single_run(name)

    def enable_single_run(self, name):
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            self.widget(i).enable_single_run(name)

    def create_new_tab(self, name='New Tab'):
        i = 1
        while name in self.tab_button_dict:
            name = f"{name} {i}"
            i += 1
        new_tab = Drop_Scroll_Area()
        self.insertTab(self.count() - 1, new_tab, name)  # Insert before the plus tab
        self.tab_button_dict[name] = []
        new_tab.order_changed.connect(self.update_order)

    def addTab(self, widget, name):
        super().addTab(widget, name)
        if name not in self.tab_button_dict and name != "":
            self.tab_button_dict[name] = []
        if name != "":
            widget.order_changed.connect(self.update_order)

    def get_active_tab(self):
        return self.tabText(self.currentIndex())

    def update_order(self):
        new_dict = OrderedDict()
        for i in range(self.count() - 1):  # Skip the last tab (plus tab)
            tab_name = self.tabText(i)
            new_dict[tab_name] = self.widget(i).get_button_order()
        self.tab_button_dict =  new_dict
        return new_dict


class MoveDialog(QDialog):
    def __init__(self, parent=None, f=Qt.WindowFlags(), button_name=""):
        super().__init__(parent, f)
        self.setWindowTitle(f"Move {button_name}")
        layout = QGridLayout()
        self.setLayout(layout)
        self.label = QLabel(f'Select the tab where to move "{button_name}"')
        self.combo = QComboBox()
        self.ok_button = QPushButton("OK")
        self.cancel_button = QPushButton("Cancel")
        self.ok_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)
        layout.addWidget(self.label, 0, 0, 1, 2)
        layout.addWidget(self.combo, 1, 0, 1, 2)
        layout.addWidget(self.ok_button, 2, 0)
        layout.addWidget(self.cancel_button, 2, 1)

    def add_tabs_from_widget(self, widget):
        for i in range(widget.count() - 1):
            self.combo.addItem(widget.tabText(i))

    def get_tab(self):
        return self.combo.currentText()


if __name__ == "__main__":
    app = QApplication([])
    main_window = QMainWindow()
    tabbing = RenameTabWidget()
    main_window.setCentralWidget(tabbing)

    area1 = Drop_Scroll_Area()
    tabbing.addTab(area1, "Tab 1")
    area2 = Drop_Scroll_Area()
    tabbing.addTab(area2, "Tab 2")
    for i in range(15):
        area1.add_button(DragButton(f"A {i}"), str(i))
        area2.add_button(DragButton(f"B {i}"), str(i))
    main_window.show()
    app.exec()



import pytest
from fastapi.testclient import TestClient
from fastapi import status, Depends, HTTPException
from unittest.mock import patch, MagicMock
from nomad_camels.api.api import FastapiThread, validate_credentials
from fastapi.security import HTTPBasicCredentials, HTTPBasic
import socket
import time
import re


# Mocked dependency to replace validate_credentials during tests
def mock_validate_credentials(credentials: HTTPBasicCredentials = Depends(HTTPBasic())):
    # Check if the provided password is "valid_api_key"
    if credentials.password == "valid_api_key":
        return "mock_api_key"
    else:
        # Raise HTTP 401 Unauthorized if the password is invalid
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid API Key"
        )


def is_port_available(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("127.0.0.1", port)) != 0


# Fixture to set up the TestClient and FastapiThread for testing
@pytest.fixture()
def client_and_thread():
    # Mock the main_window with a predefined protocols_dict
    main_window_mock = MagicMock()
    main_window_mock.protocols_dict = {"protocol1": "details1", "protocol2": "details2"}

    port = 1235
    while not is_port_available(port):
        port += 1
        print(port)
        if port > 9999:
            raise Exception("No available port found")
    # Create and start a FastapiThread instance with the mocked main_window
    thread = FastapiThread(main_window_mock, api_port=port)
    thread.start()

    # Get the FastAPI app and set up the TestClient
    app = thread.app
    client = TestClient(app)
    # Wait for the server to start
    timeout = 10  # seconds
    start_time = time.time()
    while True:
        try:
            response = client.get("/")
            if response.status_code == status.HTTP_200_OK:
                break
        except:
            pass
        if time.time() - start_time > timeout:
            raise TimeoutError("Server did not start within the timeout period")
        time.sleep(0.1)  # wait for 100ms before retrying

    # Override the validate_credentials dependency with the mock
    app.dependency_overrides[validate_credentials] = mock_validate_credentials

    # Mock the start_protocol signal
    thread.start_protocol_signal = MagicMock()

    # Provide the client and thread to the test
    yield client, thread

    # Stop the server after tests are done
    thread.stop_server()


# Test to check if the root URL redirects to the documentation page
@pytest.mark.order(-1)
def test_root_redirect(client_and_thread):
    client, thread = client_and_thread
    # Ensure the client exists
    assert client is not None, "Client is not initialized"

    # Ensure the thread exists and is running
    assert thread is not None, "Thread is not initialized"
    response = client.get("/")
    assert response.status_code == status.HTTP_200_OK
    assert str(response.url).endswith("/docs")


# Test to check if the /protocols endpoint returns the correct protocols
@pytest.mark.order(-1)
def test_get_protocols(client_and_thread):
    client, thread = client_and_thread
    # Ensure the client exists
    assert client is not None, "Client is not initialized"

    # Ensure the thread exists and is running
    assert thread is not None, "Thread is not initialized"
    response = client.get("/api/v1/protocols", auth=("user", "valid_api_key"))
    assert response.status_code == status.HTTP_200_OK
    assert response.json() == {"Protocols": ["protocol1", "protocol2"]}


# Test to check if a protocol can be run with a valid API key
@pytest.mark.order(-1)
def test_run_protocol(client_and_thread):
    client, thread = client_and_thread
    # Ensure the client exists
    assert client is not None, "Client is not initialized"

    # Ensure the thread exists and is running
    assert thread is not None, "Thread is not initialized"
    protocol_name = "protocol1"

    # Patch the emit method of the start_protocol signal
    with patch.object(thread.start_protocol_signal, "emit") as mock_emit:
        response = client.get(
            f"/api/v1/actions/run/protocols/{protocol_name}",
            auth=("user", "valid_api_key"),
        )
        assert response.status_code == status.HTTP_200_OK
        # Check if the response contains the expected URL with a UUID4 string
        response_json = response.json()
        assert "check protocol status here" in response_json
        url = response_json["check protocol status here"]
        uuid_pattern = re.compile(
            r"^/api/v1/protocols/results/[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"
        )
        assert uuid_pattern.match(url), f"URL does not match expected pattern: {url}"
        # Extract the UUID from the URL
        uuid = url.split("/")[-1]
        # Update the mock_emit assertion to include the additional parameters
        mock_emit.assert_called_once_with(protocol_name, uuid, None)


# Test to check if accessing /protocols with an invalid API key returns 401
@pytest.mark.order(-1)
def test_invalid_api_key(client_and_thread):
    client, thread = client_and_thread
    # Ensure the client exists
    assert client is not None, "Client is not initialized"

    # Ensure the thread exists and is running
    assert thread is not None, "Thread is not initialized"
    response = client.get("/api/v1/protocols", auth=("user", "invalid_api_key"))
    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert response.json() == {"detail": "Invalid API Key"}


# Test to check if the /favicon.ico endpoint returns the favicon correctly
@pytest.mark.order(-1)
def test_favicon(client_and_thread):
    client, thread = client_and_thread
    # Ensure the client exists
    assert client is not None, "Client is not initialized"

    # Ensure the thread exists and is running
    assert thread is not None, "Thread is not initialized"
    response = client.get("/favicon.ico")
    assert response.status_code == status.HTTP_200_OK
    assert response.headers["content-type"] in [
        "image/x-icon",
        "image/vnd.microsoft.icon",
    ]



import sys
import os
import requests
import re
import importlib
import importlib_metadata
import pathlib
import pkg_resources
import subprocess

sys.path.append(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
)
from nomad_camels.ui_widgets.warn_popup import WarnPopup
from nomad_camels.utility import variables_handling
from nomad_camels.main_classes.device_class import Simple_Config_Sub

from PySide6.QtWidgets import (
    QDialog,
    QTableWidget,
    QTableWidgetItem,
    QLineEdit,
    QGridLayout,
    QPushButton,
    QCheckBox,
    QSplitter,
    QTextBrowser,
    QLabel,
    QWidget,
    QDialogButtonBox,
)
from PySide6.QtCore import Qt, QThread, Signal
from PySide6.QtGui import QFont

url = "https://raw.githubusercontent.com/FAU-LAP/CAMELS_extensions/extension_list/extension_list.txt"
repo_url = "https://raw.githubusercontent.com/FAU-LAP/CAMELS_extensions/main"
camels_extension_regex = r"^(nomad[-_]{1}camels[-_]{1}extension[-_]{1})(.*)$"

bold_font = QFont()
bold_font.setBold(True)


def get_online_extensions():
    """Returns a list of all extensions."""
    online_extensions = {}
    try:
        extension_str = requests.get(url).text
        for x in extension_str.splitlines():
            if "==" not in x:
                continue
            name, version = x.split("==")
            online_extensions[name.replace("-", "_")] = version
    except:
        WarnPopup(
            None,
            "Could not (completely) read extension_list.txt from repo.\n"
            "Check internet connectivity.",
            "No extensions found online",
        )
    return online_extensions


def get_installed_extensions():
    """Returns a list of all installed extensions."""
    installed_extensions = {}
    for x in importlib_metadata.distributions():
        name = x.metadata["Name"]
        version = x.version
        if re.match(camels_extension_regex, name, re.IGNORECASE):
            installed_extensions[name[23:].replace("-", "_")] = version
    return installed_extensions


def get_local_extensions():
    """Returns a list of all local extensions."""
    local_extensions = {}
    if "extension_path" in variables_handling.preferences:
        local_extension_path = variables_handling.preferences["extension_path"]
    else:
        from nomad_camels.utility.load_save_functions import standard_pref

        local_extension_path = standard_pref["extension_path"]
    if not os.path.isdir(local_extension_path):
        return local_extensions
    sys.path.append(local_extension_path)
    for f in pathlib.Path(local_extension_path).rglob("*"):
        match = re.match(camels_extension_regex, f.name, re.IGNORECASE)
        if match:
            try:
                sys.path.append(str(f.parent))
                package = importlib.import_module(match.group(0))
                name = match.group(2)
                version = package.EXTENSION_CONFIG["version"]
                local_extensions[name] = version
            except Exception as e:
                print(f, e)
    return local_extensions


def get_readme_text(extension):
    text_url = f"{repo_url}/{extension}/README.md"
    return requests.get(text_url).text


def get_license_text(extension):
    text_url = f"{repo_url}/{extension}/LICENSE.txt"
    return requests.get(text_url).text


class Extension_Manager(QDialog):
    def __init__(self, preferences, parent=None):
        super().__init__(parent=parent)
        self.actual_preferences = preferences
        self.preferences = dict(preferences)
        self.setWindowTitle("Extension Manager - NOMAD CAMELS")
        if "extensions" not in self.preferences:
            self.preferences["extensions"] = []
        self.extensions = self.preferences["extensions"]
        if "extension_settings" not in self.preferences:
            self.preferences["extension_settings"] = {}
        self.extension_data = self.preferences["extension_settings"]

        self.online_extensions = get_online_extensions()
        self.local_extensions = get_local_extensions()
        self.installed_extensions = get_installed_extensions()

        self.extension_table = QTableWidget()
        self.extension_table.setColumnCount(3)
        self.extension_table.setHorizontalHeaderLabels(["Name", "Version", "Installed"])
        self.extension_table.verticalHeader().setVisible(False)

        self.lineEdit_search_name = QLineEdit()
        self.lineEdit_search_name.setPlaceholderText("Search...")
        self.lineEdit_search_name.textChanged.connect(self.build_table)

        self.show_hide_info_button = QPushButton("Show Info")
        self.info_hidden = True
        self.show_hide_info_button.clicked.connect(self.show_hide_info)

        self.selection_label = QLabel("Extension:")
        self.selection_label.setFont(bold_font)

        self.install_button = QPushButton("Install / Update")
        self.install_button.clicked.connect(self.install_selected)

        self.uninstall_button = QPushButton("Uninstall")
        self.uninstall_button.clicked.connect(
            lambda: self.install_selected(uninstall=True)
        )

        self.checkbox_active = QCheckBox("extension active")

        self.info_widget = Info_Widget()

        self.extension_table.clicked.connect(self.table_click)
        self.config_widget = QWidget()

        self.dialog_buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        self.dialog_buttons.accepted.connect(self.accept)
        self.dialog_buttons.rejected.connect(self.reject)

        layout = QGridLayout()
        layout.addWidget(self.lineEdit_search_name, 0, 0)
        layout.addWidget(self.extension_table, 1, 0, 5, 1)
        layout.addWidget(self.selection_label, 0, 1)
        layout.addWidget(self.install_button, 2, 1)
        layout.addWidget(self.uninstall_button, 3, 1)
        layout.addWidget(self.show_hide_info_button, 1, 1)
        layout.addWidget(self.checkbox_active, 4, 1)
        layout.addWidget(self.config_widget, 5, 1)
        layout.addWidget(self.info_widget, 0, 2, 6, 1)
        layout.addWidget(self.dialog_buttons, 6, 0, 1, 3)
        self.setLayout(layout)

        self.current_extension = None

        self.install_thread = None
        self.build_table()
        self.table_click()

    def propagate_exception(self, e):
        raise e

    def install_selected(self, uninstall=False):
        self.setEnabled(False)
        self.setCursor(Qt.WaitCursor)
        ind = self.extension_table.selectedIndexes()[0]
        ext = self.extension_table.item(ind.row(), 0).text()
        self.install_thread = Install_Thread(ext, uninstall=uninstall)
        self.install_thread.exception_signal.connect(self.propagate_exception)
        self.install_thread.finished.connect(self.build_table)
        self.install_thread.start()

    def show_hide_info(self):
        if self.info_hidden:
            self.info_widget.setHidden(False)
            self.show_hide_info_button.setText("Hide Info")
        else:
            self.info_widget.setHidden(True)
            self.show_hide_info_button.setText("Show Info")
        self.info_hidden = not self.info_hidden
        self.table_click()

    def update_config(self):
        try:
            ind = self.extension_table.selectedIndexes()[0]
        except:
            return
        ext = self.extension_table.item(ind.row(), 0).text()
        if self.current_extension:
            self.extension_data[self.current_extension] = (
                self.config_widget.get_config()
            )
        package = importlib.import_module(f"nomad_camels_extension_{ext}")
        # if checkbox checked, add to self.extensions if not already there
        name = f'nomad_camels_extension_{ext.replace("-", "_")}'
        if self.checkbox_active.isChecked():
            if name not in self.extensions:
                self.extensions.append(name)
        # if checkbox unchecked, remove from self.extensions if there
        else:
            if name in self.extensions:
                self.extensions.remove(name)
        if "settings" not in package.EXTENSION_CONFIG:
            return
        settings = package.EXTENSION_CONFIG["settings"]
        if ext in self.extension_data:
            settings.update(self.extension_data[ext])
        config_old = self.config_widget
        self.config_widget = Simple_Config_Sub(config_dict=settings)
        self.layout().replaceWidget(config_old, self.config_widget)
        config_old.deleteLater()

    def table_click(self):
        self.setCursor(Qt.WaitCursor)
        try:
            ind = self.extension_table.selectedIndexes()[0]
            ext = self.extension_table.item(ind.row(), 0).text()
            installed = ext in self.installed_extensions
            local = ext in self.local_extensions
            if local:
                self.selection_label.setText(f"Local extension: {ext}")
            else:
                self.selection_label.setText(f"Extension: {ext}")
            name = f'nomad_camels_extension_{ext.replace("-", "_")}'
            self.checkbox_active.setChecked(name in self.extensions)
            if local or installed:
                self.update_config()
            self.install_button.setEnabled(not local)
            self.uninstall_button.setEnabled(installed and not local)
            self.checkbox_active.setEnabled(installed or local)
            self.show_hide_info_button.setEnabled(True)
            if not self.info_hidden:
                self.info_widget.update_texts(ext)
            self.current_extension = ext
        except Exception as e:
            self.info_widget.setHidden(True)
            self.install_button.setEnabled(False)
            self.uninstall_button.setEnabled(False)
            self.checkbox_active.setEnabled(False)
            self.selection_label.setText("Extension:")
            self.show_hide_info_button.setEnabled(False)
            print(e)
        finally:
            self.adjustSize()
            self.setCursor(Qt.ArrowCursor)

    def build_table(self):
        self.setEnabled(True)
        self.setCursor(Qt.WaitCursor)
        search_text = self.lineEdit_search_name.text()
        self.extension_table.clear()
        self.extension_table.setRowCount(0)
        for i, ext in enumerate(sorted(self.online_extensions.keys())):
            if search_text.lower() not in ext.lower():
                continue
            self.extension_table.insertRow(i)
            item = QTableWidgetItem(ext)
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.extension_table.setItem(i, 0, item)
            item = QTableWidgetItem(self.online_extensions[ext])
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.extension_table.setItem(i, 1, item)
            if ext in self.installed_extensions:
                item = QTableWidgetItem(self.installed_extensions[ext])
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                self.extension_table.setItem(i, 2, item)
            elif ext in self.local_extensions:
                item = QTableWidgetItem(self.local_extensions[ext])
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                self.extension_table.setItem(i, 2, item)
        for i, ext in enumerate(sorted(self.local_extensions.keys())):
            if search_text.lower() not in ext.lower():
                continue
            if ext in self.online_extensions:
                continue
            self.extension_table.insertRow(i)
            item = QTableWidgetItem(ext)
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.extension_table.setItem(i, 0, item)
            item = QTableWidgetItem(self.local_extensions[ext])
            item.setFlags(item.flags() & ~Qt.ItemIsEditable)
            self.extension_table.setItem(i, 1, item)
            if ext in self.installed_extensions:
                item = QTableWidgetItem(self.installed_extensions[ext])
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                self.extension_table.setItem(i, 2, item)
        self.extension_table.resizeColumnsToContents()
        self.setCursor(Qt.ArrowCursor)

    def accept(self):
        self.update_config()
        self.actual_preferences.update(self.preferences)
        super().accept()


class Info_Widget(QSplitter):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setOrientation(Qt.Orientation.Horizontal)

        self.info_text = QTextBrowser()
        self.info_text.setOpenExternalLinks(True)
        self.info_text.setTextInteractionFlags(
            Qt.TextSelectableByKeyboard
            | Qt.TextSelectableByMouse
            | Qt.LinksAccessibleByMouse
        )
        self.addWidget(self.info_text)

        self.license_text = QTextBrowser()
        self.license_text.setOpenExternalLinks(True)
        self.license_text.setTextInteractionFlags(
            Qt.TextSelectableByKeyboard
            | Qt.TextSelectableByMouse
            | Qt.LinksAccessibleByMouse
        )
        self.addWidget(self.license_text)
        self.info = False

        self.setContentsMargins(0, 0, 0, 0)

    def update_texts(self, extension):
        self.info = False
        hidden = True
        try:
            self.info_text.setHidden(True)
            self.info_text.clear()
            try:
                meta = importlib_metadata.metadata(
                    f"nomad_camels_extension_{extension}"
                )
                text = meta.json["description"]
            except:
                text = get_readme_text(extension)
            self.info_text.setMarkdown(text)
            self.info_text.setHidden(False)
            hidden = False
        except:
            pass
        try:
            self.license_text.setHidden(True)
            self.license_text.clear()
            try:
                text = ""
                for p in pkg_resources.working_set:
                    if not p.key.startswith(
                        f'nomad-camels-driver-{extension.replace("_", "-")}'
                    ):
                        continue
                    lic = p.get_metadata_lines("LICENSE.txt")
                    for l in lic:
                        text += f"{l}\n"
                    break
                if not text:
                    raise Exception("")
                self.license_text.setText(text)
            except:
                text = get_license_text(extension)
                self.license_text.setMarkdown(text)
            self.license_text.setHidden(False)
            hidden = False
        except:
            pass
        self.info = not hidden
        self.setHidden(hidden)


class Install_Thread(QThread):
    exception_signal = Signal(Exception)

    def __init__(self, extension, uninstall=False, parent=None):
        super().__init__(parent=parent)
        self.extension = extension
        self.uninstall = uninstall

    def run(self):
        try:
            name = self.extension.replace("_", "-")
            flags = 0
            if os.name == "nt":
                flags = subprocess.CREATE_NO_WINDOW
            if self.uninstall:
                ret = subprocess.Popen(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "uninstall",
                        "-y",
                        f"nomad-camels-extension-{name}",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    creationflags=flags,
                )
            else:
                ret = subprocess.Popen(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        f"nomad-camels-extension-{name}",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    creationflags=flags,
                )
            for line in iter(ret.stdout.readline, b""):
                text = line.decode().rstrip()
                print(text)
            errs = ret.stderr.read().decode()
            if ret.returncode or errs:
                self.exception_signal.emit(Exception(errs))
        except Exception as e:
            self.exception_signal.emit(e)


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication

    app = QApplication([])

    widge = Extension_Manager({})
    widge.show()
    app.exec()
